// Mandatory unit header per GUI SAC, Section 6.1.1
//	Package: Provider Role Tool GUI application
//	Date Created: August 2017
//  Site Name: CPRS EP1 Remote Development Team
// Developers: Scott Stephenson (Delphi), Chris Bell (Delphi)
//             Robert Davis (M), Jamie Crumley (M)
//	Description: Review and confirmation form - last step before calling RPC
//	Note: This application requires OR_3_453 in order to run.

unit fApplyChanges;

interface

uses
  uProviderEngine,
  fBase508Form,
  ORDtTm,
  VA508AccessibilityManager,
  ORCtrls, uUtilities,
  // standard Delphi units
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ExtCtrls, System.Actions,
  Vcl.ActnList, Vcl.StdCtrls, Vcl.ComCtrls,
  Vcl.Menus, Clipbrd,
  //Needs ot be last for the button extension
  U508Extensions;

type

  TApplyChangesForm = class(TfrmBase508Form)
    pnlBottom: TPanel;
    btnClose: TButton;
    btnExecute: TButton;
    actionsMain: TActionList;
    ActionClose: TAction;
    ActionExecute: TAction;
    pnlTop: TPanel;
    ORDTEffectiveTime: TORDateBox;
    lbl508StatusPrompt: TVA508StaticText;
    ActionCopyToClipboard: TAction;
    pmResults: TPopupMenu;
    popCopyResultstoClipboard: TMenuItem;
    lblEffectiveDateTime: TLabel;
    lbl508HintDateTime: TLabel;
    lvOrders: TListView;
    procedure ActionCloseExecute(Sender: TObject);
    procedure ActionExecuteExecute(Sender: TObject);
    procedure actionsMainUpdate(Action: TBasicAction; var Handled: Boolean);
    procedure lvOrdersCustomDrawSubItem(Sender: TCustomListView;
      Item: TListItem; SubItem: Integer; State: TCustomDrawState;
      var DefaultDraw: Boolean);
    procedure ActionCopyToClipboardExecute(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ORDTEffectiveTimeChange(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    fOrdersReassigned: boolean;
    fOrderEngine: TOrderEngine;
    fReassignedOrders: tReassignOrderRecords;
    fLV508Manager: Tlv508Manager;
    procedure ShowOrders;
  public
    Property OrderReassigned: Boolean read fOrdersReassigned;
  end;

function ReviewAndApplyChanges(OrderEngine: TOrderEngine; reassignedOrders: tReassignOrderRecords): boolean;

implementation

{$R *.dfm}

uses
  ORFn,
  DateUtils,
  VA508AccessibilityRouter,
  System.Generics.Collections,
  System.Generics.Defaults,
  uSizeEngine,
  vaUtils;

const
  // IDxxx are the treelist column indexes
  IDPatient = 0;
  IDOldProvider = 1;
  IDNewProvider = 2;
  IDResult = 3;
  IDError = 4;
  SFail = 'FAIL';
  SPass = 'OK';
  MSG508ReportTxt = 'You can obtain a text record of these transfers using the context menu.';
  MSGReportTxt = 'You can obtain a text record of these transfers using the pop up menu by right clicking in the results window.';

function ReviewAndApplyChanges(OrderEngine: TOrderEngine; ReassignedOrders: tReassignOrderRecords): boolean;
var
  f: TApplyChangesForm;
begin
  f:= TApplyChangesForm.Create(Application);
  try
    //Set to relative date/time
    f.ORDTEffectiveTime.FMDateTime := DateTimeToFMDateTime(now);
    f.fReassignedOrders:= ReassignedOrders;
    f.FOrderEngine:= OrderEngine;
    f.ShowOrders;
    f.ShowModal;
   // result:= f.ShowModal = MROK;
    Result := f.OrderReassigned;
  finally
    f.Release;
  end;
end;


procedure TApplyChangesForm.ActionCloseExecute(Sender: TObject);
// Close the form without doing anything
begin
  Close;
  if fOrdersReassigned then
   modalResult:= mrOK
  else
    modalResult:= mrCancel;
end;


procedure TApplyChangesForm.ActionCopyToClipboardExecute(Sender: TObject);
// copy the results to clipboard
var
  i, aMaxLine, aMaxCol2, aMaxCol3, aMaxCol4, aMaxCol5: integer;
  s: String;
  AItem: TListItem;
  RtnLst: TStringList;
  aSpaces: String;
begin
  RtnLst := TStringList.Create;
  try
    //Figure out the longest columns
    aMaxCol2 := 0;
    aMaxCol3 := 0;
    aMaxCol4 := 0;
    aMaxCol5 := 0;
    for AItem in lvOrders.Items do
    begin
     if Length(AItem.Caption) > aMaxCol2 then
      aMaxCol2 := Length(AItem.Caption);

      if Length(AItem.SubItems[IDPatient]) > aMaxCol3 then
      aMaxCol3 := Length(AItem.SubItems[IDPatient]);

      if Length(AItem.SubItems[IDOldProvider]) > aMaxCol4 then
      aMaxCol4 := Length(AItem.SubItems[IDOldProvider]);

      if Length(AItem.SubItems[IDNewProvider]) > aMaxCol5 then
      aMaxCol5 := Length(AItem.SubItems[IDNewProvider]);
    end;
    aSpaces := Format('%4s', [' ']);
    //aSpaces := '     ';
    //Add headers
    RtnLst.Add(
      format('%3s '+aSpaces+
      ' %'+IntToStr(aMaxCol2)+'s '+ aSpaces+
      ' %'+IntToStr(aMaxCol3)+'s '+ aSpaces+
      ' %'+IntToStr(aMaxCol4)+'s '+ aSpaces+
      ' %'+IntToStr(aMaxCol5)+'s '+ aSpaces+
      ' %7s '+ aSpaces+
      ' %s',
      ['ID',
       'Order #',
       'Patient',
       'Old Provider',
       'New Provider',
       'Results',
       'Error Message']));

    RtnLst.Add('');

    //Add individual items
    for i:= 0 to lvOrders.Items.Count - 1 do
    begin
      AItem:= lvOrders.Items[i];
      RtnLst.Add(format('%3d. '+ aSpaces+' %'+IntToStr(aMaxCol2)+'s '+ aSpaces+' %'+IntToStr(aMaxCol3)+'s '+ aSpaces+' %'+IntToStr(aMaxCol4)+'s '+ aSpaces+' %'+IntToStr(aMaxCol5)+'s '+ aSpaces+' %7s '+ aSpaces+' %s',
        [
        i + 1,
        AItem.Caption, // order IEN
        AItem.SubItems[IDPatient],
        AItem.SubItems[IDOldProvider],
        AItem.SubItems[IDNewProvider],
        AItem.SubItems[IDResult],
        AItem.SubItems[IDError]
        ]));
    end;

    //Need to figure out the longest line
    aMaxLine := 0;
    for I := 0 to RtnLst.Count -1 do
    begin
     if RtnLst.Strings[i].Length > aMaxLine then
      aMaxLine := RtnLst.Strings[i].Length;
    end;
    RtnLst.Insert(1, StringOfChar('=', aMaxLine));

    Clipboard.AsText:= RtnLst.Text;
  finally
    RtnLst.Free;
  end;
end;


procedure TApplyChangesForm.ActionExecuteExecute(Sender: TObject);
Const
  userAdvisory =
    'Order transfer %s%s. Please carefully review the results and ' +
    'remember to check all providers for possible surrogates.' + SLineBreak + SLineBreak;
  ErrorMsg = ' with %s error(s) occurring during the transfer' ;
var
  i: integer;
  strEffectiveTransferDateTime, MsgStr, HdrTxt: string;
  ErrCnt: Integer;
  dlgTyp: tTaskDLGType;
begin
  //Validate we have a good time
  if ORDTEffectiveTime.FMDateTime < DateTimeToFMDateTime(Date) then
  begin
    InfoBox('You must enter current or a future date', 'Invalid date', MB_OK or MB_ICONERROR);
    exit;
  end;

  fOrdersReassigned:= true; // must be first because we want the assignment button disabled in the action update
  ActionsMain.UpdateAction(nil);

  strEffectiveTransferDateTime := floatToStr(ORDTEffectiveTime.FMDateTime);
  for i := 0 to length(FReassignedOrders) - 1 do
    FReassignedOrders[i].effectiveDateTime:= strEffectiveTransferDateTime;

  // perform the actual reassignment on the VistA server
  fOrderEngine.reassignOrders(fReassignedOrders);

  // update the order display to show any failures
  ShowOrders;
  // change the caption of btnClose (actionClose) from "Cancel" to "Close" for clarity
  ActionClose.Caption:= '&Close';

  ErrCnt := 0;
  for i := 0 to length(FReassignedOrders) - 1 do
    if Trim(FReassignedOrders[i].errorMessage) <> '' then
      Inc(ErrCnt);

  // show a popup message to make it really obvious we've reassigned orders
  if (ErrCnt > 0) and (ErrCnt <> Length(FReassignedOrders)) then
  begin
    dlgTyp := tdlgWarning;
    HdrTxt := 'Transfer Partially Completed';
    MsgStr := format(userAdvisory, ['partially completed', format(ErrorMsg, [IntToStr(ErrCnt)])]);
  end else if ErrCnt = Length(FReassignedOrders)  then
  begin
    dlgTyp := tdlgError;
    HdrTxt := 'Transfer Failed';
    MsgStr := format(userAdvisory, ['failed', format(ErrorMsg, [IntToStr(ErrCnt)])]);
  end else
  begin
    dlgTyp := tdlgInformation;
    HdrTxt := 'Transfer Completed';
    MsgStr := format(userAdvisory, ['completed', '']);
  end;

  if ScreenReaderSystemActive then
     MsgStr:= MsgStr + MSG508ReportTxt
  else
    MsgStr:= MsgStr + MSGReportTxt;
 // InfoBox(userAdvisory, 'Transfer Complete', MB_OK);
  PRTMessage('Provider Role Tool', HdrTxt, MsgStr, ['OK'], 'OK', dlgTyp);
end;


procedure TApplyChangesForm.actionsMainUpdate(Action: TBasicAction; var Handled: Boolean);
// set control state
begin
  actionClose.Enabled:= not FOrderEngine.working;
  actionExecute.Enabled:= not fOrdersReassigned;
  ORDTEffectiveTime.Enabled:= not fOrdersReassigned;
end;

{constructor TApplyChangesForm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;   }

procedure TApplyChangesForm.FormCreate(Sender: TObject);
begin
  inherited;
  // set default values
  fOrdersReassigned:= false;
  // enforce visibility of the orders grid on resize
 // lvOrders.Constraints.MinHeight:= 100;
 // lvOrders.Constraints.MinWidth:= 200;
  self.Constraints.MinWidth:= ORDTEffectiveTime.Left + ORDTEffectiveTime.Width + 10;
  ORDTEffectiveTime.DateRange.MinDate := DateOf(now);
  ORDTEffectiveTime.DateRange.MaxDate := Now + 7;

  SizeHolder.LoadSize(self);
  SizeHolder.LoadSize(lvOrders);
  if ScreenReaderActive then
  begin
    fLV508Manager := Tlv508Manager.Create;
    amgrMain.ComponentManager[lvOrders] := fLV508Manager;
  end;
end;

procedure TApplyChangesForm.FormDestroy(Sender: TObject);
begin
  SizeHolder.SetSize(self);
  SizeHolder.SetSize(lvOrders);

  if assigned(fLV508Manager) then
  begin
    amgrMain.ComponentManager[lvOrders] := nil;
    FreeAndNil(fLV508Manager);
  end;

  inherited;
end;

procedure TApplyChangesForm.lvOrdersCustomDrawSubItem(Sender: TCustomListView;
  Item: TListItem; SubItem: Integer; State: TCustomDrawState;
  var DefaultDraw: Boolean);
begin
  if (subItem = IDError) then
  begin
    if (Item.SubItems[IDError] = SFail) then
      lvOrders.Canvas.Brush.color:= clYellow
    else if (Item.SubItems[IDError] = SPass) then
      lvOrders.Canvas.Brush.color:= clMoneyGreen
    else
      lvOrders.Canvas.Brush.color:= clWhite;
  end;
end;


procedure TApplyChangesForm.ORDTEffectiveTimeChange(Sender: TObject);
var
  ErrMsg: String;
begin
  inherited;
  if IsFMDateTime(FloatToStr(ORDTEffectiveTime.FMDateTime)) then
  begin
    ORDTEffectiveTime.Validate(ErrMsg);
    if length(ErrMsg) > 0 then
    begin
      PRTMessage(ORDTEffectiveTime.Caption, ORDTEffectiveTime.Caption, ErrMsg, ['OK'], 'OK', tdlgError);
      ActionExecute.Enabled := false;
      lvOrders.Enabled := false;
    end else
    begin
      ActionExecute.Enabled := true;
      lvOrders.Enabled := true;
    end;
  end else begin
  ActionExecute.Enabled := False;
  lvOrders.Enabled := False;
  end;
end;

procedure TApplyChangesForm.ShowOrders;
// display orders before we try reassignment
// reuse this method to show results after reassignment

  procedure SortByNewProviderIEN;
  begin
    TArray.Sort<tReassignOrderRecord>(FReassignedOrders , TDelegatedComparer<tReassignOrderRecord>.Construct(
    function(const Left, Right: tReassignOrderRecord): Integer
    begin
      Result := TComparer<String>.Default.Compare(Left.newProviderIEN, Right.newProviderIEN);
    end));
  end;

var
  aOrder: TReassignOrderRecord;
  aItem: TListItem;
  numFailed, gpID: integer;
  fLastGroup: String;
  Group: TListGroup;
begin
  numFailed:= 0;
  lvOrders.Items.BeginUpdate;
  try
    lvOrders.Items.Clear;
    SortByNewProviderIEN;
    LVOrders.GroupView := true;
    gpID := 0;
    fLastGroup := '';
    Group := nil;
    for aOrder in FReassignedOrders do
    begin
      aItem:= LVOrders.Items.Add;

      if aOrder.newProviderIEN <> fLastGroup then
      begin
        Inc(gpID);
        Group := LVOrders.Groups.Add;
        Group.GroupID := gpID;
        Group.Header := aOrder.newProviderName;
        fLastGroup := aOrder.newProviderIEN;
      end;

      if assigned(Group) then
        aItem.GroupID := Group.GroupID;
      aItem.Caption:= aOrder.OrderIEN;
      aItem.SubItems.Add(format('(%s) %s', [aOrder.patientIEN, aOrder.patientName]));
      aItem.SubItems.Add(format('(%s) %s', [aOrder.oldProviderIEN, aOrder.OldProviderName]));
      aItem.SubItems.Add(format('(%s) %s', [aOrder.newProviderIEN, aOrder.NewProviderName]));
      if aOrder.result = '0' then
      begin
        aItem.SubItems.Add(SFail);
        inc(numFailed);
      end
      else if aOrder.result = '1' then
        aItem.SubItems.Add(SPass)
      else
        AItem.SubItems.Add('');
      aItem.SubItems.Add(aOrder.ErrorMessage);

    end;
  finally
    lvOrders.Items.EndUpdate;
  end;

  // show a hint below the orders list view
  if fOrdersReassigned then
    lbl508StatusPrompt.caption:= format(
      'Results: %d succeeded. %d failed. Click "Close" to exit',
      [lvOrders.Items.Count - numFailed, numFailed])
  else
    lbl508StatusPrompt.caption:= format(
    '%d pending order reassignments. Click "Reassign Orders" to send to server.',
    [lvOrders.Items.Count - numFailed, numFailed]);

end;

end.
