// Mandatory unit header per GUI SAC, Section 6.1.1
// Package: Provider Role Tool GUI application
// Date Created: Janurary 2020
// Site Name: CPRS EP1 Remote Development Team
// Developers: Chris Bell (Delphi) Jamie Crumley (M)
// Description: Audit screen
// Note: This application requires OR*3.0*453 in order to run.

unit fAudit;

interface

uses
  ORCtrls, VA508AccessibilityManager, ORDtTm, uUtilities, ffilter,
  // standard Delphi units
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants,
  System.Classes, Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs,
  Vcl.Menus, Vcl.ComCtrls, Vcl.StdCtrls, Vcl.ExtCtrls,
  System.ImageList, System.Actions, Vcl.ActnList, Vcl.ImgList,
  //Needs ot be last for the button extension
  U508Extensions;

type
  // Fitler action
  tFilterAction = (fl_Sort, fl_filter);

  // Holds the filter forms
  tFilterRecord = record
    ColumnIndex: Integer;
    FilterForm: TFrmFilter;
  end;

  tFilterArray = array of tFilterRecord;

  // Helper for tFilterArray
  TFilterArrayHelper = record helper for tFilterArray
  private
    function FindRecord(const ColIdx: Integer): Integer;
    function GetFilterForm(const ColIdx: Integer): TFrmFilter;
    function AddFilter(const ColIdx: Integer): Integer;
  public
    procedure ClearList;
    property FilterFrom[const ColIdx: Integer]: TFrmFilter read GetFilterForm;
  end;

  TfrmAudit = class(TForm)
    pnlResults: TPanel;
    pnlCritera: TPanel;
    pnlBottom: TPanel;
    btnOK: TButton;
    grpbxCriteria: TGroupBox;
    pnlCritLeft: TPanel;
    gpnlCriteria: TGridPanel;
    pnlStart: TPanel;
    pnlStop: TPanel;
    lblStartDate: TLabel;
    ordbStart: TORDateBox;
    ordbStop: TORDateBox;
    LblStopDate: TLabel;
    lblResults: TVA508StaticText;
    btnRunReport: TButton;
    lvResults: TListView;
    ImageList1: TImageList;
    btnShowOrders: TButton;
    ActionList1: TActionList;
    actShowOrders: TAction;
    actBP: TAction;
    amgrMain: TVA508AccessibilityManager;
    gbColOpt: TGroupBox;
    cmbColumn: TComboBox;
    Label1: TLabel;
    Button1: TButton;
    btnSort: TButton;
    actSort: TAction;
    actFilter: TAction;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure lvResultsColumnClick(Sender: TObject; Column: TListColumn);
    procedure lvResultsColumnRightClick(Sender: TObject; Column: TListColumn;
      Point: TPoint);
    procedure lvResultsData(Sender: TObject; Item: TListItem);
    procedure FormResize(Sender: TObject);
    procedure actShowOrdersUpdate(Sender: TObject);
    procedure actShowOrdersExecute(Sender: TObject);
    procedure actBPExecute(Sender: TObject);
    procedure actSortExecute(Sender: TObject);
    procedure actFilterExecute(Sender: TObject);
    procedure ordbStartChange(Sender: TObject);
    procedure ordbStopChange(Sender: TObject);
  private
    { Private declarations }
    fDateStart: tDate;
    fDateStop: tDate;
    fPSB: tProgressStatusBar;
    fLoading: Boolean;
    fFilterArray: tFilterArray;
    fMasterList: TStringList;
    fFilterList: TStringList;
    fLV508Manager: Tlv508Manager;
    procedure AddColumnFilters;
    procedure SetColumnWidth();
    procedure AdjustFilter(const Value: String; const ColumnNum: Integer;
      Const AddToFilter: Boolean);
    procedure DrawColumnImage(Column: TListColumn; FLAction: tFilterAction);
    procedure ValidateCriteria;
  public
    { Public declarations }
    property AuditList: TStringList read fFilterList;
  end;

function RunAuditReport(): Boolean;

var
  Descending: Boolean;
  SortColumn: Integer;
  frmAudit: TfrmAudit;

Const
  //Image index
  Idx_Filter = 0;
  Idx_FilterDown = 1;
  Idx_FilterUp = 2;
  Idx_down = 3;
  Idx_Up = 4;

implementation

uses
  DateUtils, ORFN, System.Math, uSizeEngine, Winapi.CommCtrl, fPatientOrders,
  fReassignOrders, system.uitypes, VAUtils;

function RunAuditReport(): Boolean;
// ********************************************************
// Setup and run the Audit Report
// ********************************************************
begin
  frmAudit := TfrmAudit.Create(Application);
  try
    result := frmAudit.ShowModal = MROK;
  finally
    frmAudit.Release;
  end;
end;

{$R *.dfm}

procedure TfrmAudit.FormCreate(Sender: TObject);
// ********************************************************
// Setup the Audit Report with defaults
// ********************************************************
var
  I: integer;
begin
  inherited;
  fLoading := true;
  SizeHolder.LoadSize(self);
  fDateStop := Date;
  fDateStart := IncYear(fDateStop, -1);
  ordbStart.DateSelected := fDateStart;
  ordbStop.DateSelected := fDateStop;
  ordbStart.FMDateTime := DateTimeToFMDateTime(fDateStart);
  ordbStop.FMDateTime := DateTimeToFMDateTime(fDateStop);
  ordbStart.DateRange.MaxDate := FMDateTimeToDateTime(FMNow);
  ordbStop.DateRange.MaxDate := FMDateTimeToDateTime(FMNow);


  fPSB := tProgressStatusBar.Create(self);
  fPSB.Visible := true;
  fPSB.DisplayHints := true;
  fLoading := false;

  //Setup the list
  fMasterList := TStringList.Create;
  fFilterList := TStringList.Create;

  //Default sort column
  SortColumn := -1;
  if ScreenReaderActive then
  begin
    fLV508Manager := Tlv508Manager.Create;
    amgrMain.ComponentManager[lvResults] := fLV508Manager;
  end;

  //If screen reader is on then give them a way to sort and filter
  if ScreenReaderActive then
  begin
    gbColOpt.Visible := true;
    for i := 0 to lvResults.Columns.Count - 1 do
      cmbColumn.AddItem(lvResults.Columns[i].Caption, lvResults.Columns[i]);
  end;
end;

procedure TfrmAudit.FormDestroy(Sender: TObject);
// ********************************************************
// Clean up
// ********************************************************
begin
  if Assigned(fLV508Manager) then
  begin
    amgrMain.ComponentManager[lvResults] := nil;
    FreeAndNil(fLV508Manager);
  end;
  SizeHolder.SetSize(self);
  fFilterArray.ClearList;
  FreeAndNil(fMasterList);
  FreeAndNil(fFilterList);
  FreeAndNil(fPSB);
  inherited;
end;

procedure TfrmAudit.FormResize(Sender: TObject);
// ********************************************************
// Keep widths of status bar panels
// ********************************************************
begin
  if assigned(fPSB) then
    fPSB.AdjustPanelWidth(0, (frmAudit.Width - btnShowOrders.Left)*-1);
end;

procedure TfrmAudit.lvResultsColumnClick(Sender: TObject; Column: TListColumn);
// ********************************************************
// Sort the column
// ********************************************************
begin
  // Reset the prior sortcolumn
  DrawColumnImage(Column, fl_Sort);

  //Determine which column is sorting and how
  if Column.Index <> SortColumn then
  begin
    SortColumn := Column.Index;
    Descending := false;
  end
  else
    Descending := not Descending;

  //Sort the stringlist (virtual listview)
  if not Descending then
    SortByPiece(fFilterList, U, Column.Index + 1, DIR_FRWRD)
  else
    SortByPiece(fFilterList, U, Column.Index + 1, DIR_BKWRD);

  //Update the image on the sort column
  DrawColumnImage(Column, fl_Sort);
end;

procedure TfrmAudit.DrawColumnImage(Column: TListColumn;
  FLAction: tFilterAction);
// **********************************************************
// Set the image index for the column to indicate the action
// **********************************************************
begin
  // Its sorted
  if SortColumn <> -1 then
  begin
    // This  action is not the sort column
    if Column.Index <> SortColumn then
    begin
      // We must be setting a new sort column so clear the image or set the
      // filter icon
      if FLAction = fl_Sort then
      begin
        // If this old sort column is filtered then show the filtered icon
        // else clear it
        if lvResults.Columns[SortColumn].Tag = 0 then
         lvResults.Columns[SortColumn].ImageIndex := -1
        else
          lvResults.Columns[SortColumn].ImageIndex :=  Idx_Filter;
      end
      else
      begin
        // if this colum is filtered then show the filtered icon
        // else clear it
        if lvResults.Columns[Column.Index].Tag = 0 then
          lvResults.Columns[Column.Index].ImageIndex := -1
        else
          lvResults.Columns[Column.Index].ImageIndex := Idx_Filter;
      end;
    end
    else
    begin
      // We are performing a sort. Indicate that we are filtered as well
      // if applicable
      if not Descending then
      begin
        if Column.Tag = 1 then
         lvResults.Columns[SortColumn].ImageIndex := Idx_FilterDown
        else
         lvResults.Columns[SortColumn].ImageIndex := Idx_down;
      end
      else
      begin
        if Column.Tag = 1 then
         lvResults.Columns[SortColumn].ImageIndex :=  Idx_FilterUp
        else
         lvResults.Columns[SortColumn].ImageIndex :=  Idx_Up;
      end;
    end;
  end
  else
  begin
    // No sorting just filtering or nothing
    if lvResults.Columns[Column.Index].Tag = 0 then
     lvResults.Columns[Column.Index].ImageIndex := -1
    else
     lvResults.Columns[Column.Index].ImageIndex := Idx_Filter;
  end;
  //Redraw the listview
  lvResults.Invalidate;
end;

procedure TfrmAudit.actBPExecute(Sender: TObject);
// ********************************************************
// Build out the audit report
// ********************************************************
var
  UsrData: TStringList;
  rtnCsr: TCursor;
begin
  rtnCsr := screen.Cursor;
  screen.Cursor := crHourGlass;
  try
    UsrData := TStringList.Create;
    try
      lvResults.Enabled := true;
      lvResults.Clear;

      //Call to get the audit data
      AuditData(FloatToStr(ordbStart.FMDateTime),
        FloatToStr(ordbStop.FMDateTime), UsrData);

      //Sort the list by the first Name and Order IEN
      SortByPieces(UsrData, U, [1, 2]);

      //Put the results in our main list and the filter list
      fMasterList.Assign(UsrData);
      fFilterList.Assign(UsrData);

      //Updaing the display
      lvResults.Items.BeginUpdate;
      try
        //Set the displays count (virtual listview)
        lvResults.Items.Count := fMasterList.Count;

        // create the filters
        AddColumnFilters;

        //Set the column widths based on the list
        SetColumnWidth;
      finally
        lvResults.Items.EndUpdate;
      end;

      //Set the current sort coulmn and indicate
      SortColumn := 0;
      Descending := false;
      DrawColumnImage(lvResults.Columns[0], fl_Sort);

    finally
      UsrData.Free;
    end;
  finally
    screen.Cursor := rtnCsr;
  end;
end;

procedure TfrmAudit.actFilterExecute(Sender: TObject);
// ********************************************************
// 508 filtering
// ********************************************************
begin
  if assigned(cmbColumn.Items.Objects[cmbColumn.ItemIndex]) then
    lvResultsColumnRightClick(sender, tListColumn(cmbColumn.Items.Objects[cmbColumn.ItemIndex]), lvResults.ScreenToClient(Mouse.CursorPos));
end;

procedure TfrmAudit.actShowOrdersExecute(Sender: TObject);
// ********************************************************
// Dipslay details about the selected patient orders
// ********************************************************
begin
 ShowPatientOrders(nil, lvResults);
end;

procedure TfrmAudit.actShowOrdersUpdate(Sender: TObject);
// ********************************************************
// If orders are selected show the report button
// ********************************************************
begin
 actShowOrders.Enabled := lvResults.SelCount > 0;
end;

procedure TfrmAudit.actSortExecute(Sender: TObject);
// ********************************************************
// 508 sorting
// ********************************************************
begin
 if assigned(cmbColumn.Items.Objects[cmbColumn.ItemIndex]) then
  lvResultsColumnClick(sender, tListColumn(cmbColumn.Items.Objects[cmbColumn.ItemIndex]));
end;

procedure TfrmAudit.AddColumnFilters;
// ********************************************************
// Adds column data to a filters objects
// ********************************************************
var
  X, MaxCols: Integer;
  s, AddStr: String;
begin
  // We must have data to show
  if fFilterList.Count = 0 then
    exit;
  //How many columns are we going to sort on
  MaxCols := DelimCount(fFilterList[0], U);

  //Clear the forms
  for X := 1 to MaxCols do
  begin
    fFilterArray.FilterFrom[X - 1].Clear(lvResults.Column[x - 1].tag = 0);
  end;

  // Loop through each row
  for s in fFilterList do
  begin
    //Each column on this row
    for X := 1 to MaxCols do
    begin
      AddStr := Piece(s, U, X);
      if X = 1 then
        fFilterArray.FilterFrom[X - 1].AddMultiPeiceToFilters(AddStr, ';', 1, lvResults.Column[x - 1].tag = 1)
      else if X = 2 then
        fFilterArray.FilterFrom[X - 1].AddToFilters(AddStr, tFTInteger, lvResults.Column[x - 1].tag = 1)
      else if (X = 3) or (X = 5) or (X = 9) then
        fFilterArray.FilterFrom[X - 1].AddToFilters(FormatFMDateTimeStr('mm/dd/yyyy hh:nn:ss', AddStr), tFTFMDate, lvResults.Column[x - 1].tag = 1)
      else
        fFilterArray.FilterFrom[X - 1].AddToFilters(AddStr, tFTString, lvResults.Column[x - 1].tag = 1);
    end;
  end;
end;

procedure TfrmAudit.lvResultsColumnRightClick(Sender: TObject;
  Column: TListColumn; Point: TPoint);
// ********************************************************
// Display filter box
// ********************************************************
var
  h: THandle;
  r, b: TRect;
  I: Integer;
  aFrm: TFrmFilter;
begin
  // figure out where to display this box based in if it came from 508
  // filtering or not
  if Sender = actFilter then
  begin
    GetWindowRect(gbColOpt.Handle, r);
  end
  else
  begin
    h := Listview_GetHeader(lvResults.Handle);
    GetWindowRect(h, r);

    for I := 0 to lvResults.Columns.Count - 1 do
    begin
      if I >= Column.Index then
        break
      else
        Inc(r.Left, ListView_GetColumnWidth(lvResults.Handle,
          lvResults.Columns[I].Index));
    end;
  end;
  // Grab the filter form based on this column
  aFrm := fFilterArray.GetFilterForm(Column.Index);

  b := frmAudit.BoundsRect;
  if (r.Left + aFrm.Width) > b.Right then
    r.Left := (b.Right - aFrm.Width) - 10;

  //Display the filter
  if aFrm.ShowForm(r.Left, r.Bottom) = MROK then
  begin
    // rebuild the list
    for I := Low(aFrm.FilterResults) to High(aFrm.FilterResults) do
    begin
      if aFrm.FilterResults[I].Action = tfaAdd then
        AdjustFilter(aFrm.FilterResults[I].Value, Column.Index, true)
      else if aFrm.FilterResults[I].Action = tfaRemove then
        AdjustFilter(aFrm.FilterResults[I].Value, Column.Index, false);
    end;

    // Indicate if the column is filtered or not
    if aFrm.IsFiltered then
      Column.Tag := 1
    else
      Column.Tag := 0;

    // Update all the filtered items based on our new filters
    if length(aFrm.FilterResults) > 0 then
      AddColumnFilters;
    aFrm.FilterResults.Clear;

    // Update the column image
    DrawColumnImage(Column, fl_filter);

    // If sorted make sure we resort the stringlist
    if SortColumn <> -1 then
    begin
      if not Descending then
        SortByPiece(fFilterList, U, SortColumn + 1, DIR_FRWRD)
      else
        SortByPiece(fFilterList, U, SortColumn + 1, DIR_BKWRD);
    end;

  end;
  // Update the virtual ListView
  lvResults.Items.Count := fFilterList.Count;
  lvResults.Invalidate;
end;

procedure TfrmAudit.lvResultsData(Sender: TObject; Item: TListItem);
// ********************************************************
// Build the list item (virtual ListView)
// ********************************************************
var
  s: String;
begin
  s := fFilterList.Strings[Item.Index];
  Item.Caption := Piece(Piece(s, U, 1), ';', 1);
  Item.SubItems.Add(Piece(s, U, 2));
  Item.SubItems.Add(FormatDateTime('mm/dd/yyyy hh:nn:ss',
    FMDateTimeToDateTime(StrToFloatDef(Piece(s, U, 3), -1))));
  Item.SubItems.Add(Piece(s, U, 4));
  Item.SubItems.Add(FormatDateTime('mm/dd/yyyy',
    FMDateTimeToDateTime(StrToFloatDef(Piece(s, U, 5), -1))));
  Item.SubItems.Add(Piece(s, U, 6));
  Item.SubItems.Add(Piece(s, U, 7));
  Item.SubItems.Add(Piece(s, U, 8));
  if Piece(s, U, 9) <> '' then
    Item.SubItems.Add(FormatDateTime('mm/dd/yyyy hh:nn:ss',
      FMDateTimeToDateTime(StrToFloatDef(Piece(s, U, 9), -1))));
end;

procedure TfrmAudit.ordbStartChange(Sender: TObject);
// ********************************************************
// ensure date rules are followed
// ********************************************************
var
  ErrMsg: String;
begin
  if IsFMDateTime(FloatToStr(ordbStart.FMDateTime)) then
  begin
    ordbStart.Validate(ErrMsg);
    if length(ErrMsg) > 0 then
    begin
      PRTMessage(ordbStart.Caption, ordbStart.Caption, ErrMsg, ['OK'], 'OK', tdlgError);
    end else begin
      // Stop date max is today and min is start date
      ordbStop.DateRange.MinDate := FMDateTimeToDateTime(ordbStart.FMDateTime);
    end;
  end;

  ValidateCriteria;

end;

procedure TfrmAudit.ordbStopChange(Sender: TObject);
// ********************************************************
// ensure date rules are followed
// ********************************************************
var
  ErrMsg: String;
begin
   if IsFMDateTime(FloatToStr(ordbStop.FMDateTime)) then
  begin
    ordbStop.Validate(ErrMsg);
    if length(ErrMsg) > 0 then
    begin
      PRTMessage(ordbStop.Caption, ordbStop.Caption, ErrMsg, ['OK'], 'OK', tdlgError);
    end else begin
      // Start date max is today
      ordbStart.DateRange.MaxDate := FMDateTimeToDateTime(ordbStop.FMDateTime);
    end;
  end;
  ValidateCriteria;

end;

procedure TfrmAudit.ValidateCriteria;
// ********************************************************
// Validate the dates before running the report
// ********************************************************
begin
  actBP.Enabled := ordbStart.IsValid and ordbStop.IsValid;
end;

procedure TfrmAudit.AdjustFilter(const Value: String; const ColumnNum: Integer;
  Const AddToFilter: Boolean);
// ********************************************************
// adjust the filtered stringlist based on the filters
// ********************************************************

  Function CanAddWithFilters(aStr: String; FromCol: Integer): Boolean;
  // ********************************************************
  // Can we add this item based on the other filters
  // ********************************************************
  var
    I: Integer;
    ValArry: tStrArray;
    aFrm: TFrmFilter;
    s: String;
    fnd: Boolean;
  begin
    Result := true;
    fnd := false;
    //Look for the filtered column(s)
    for I := 0 to lvResults.Columns.Count - 1 do
    begin
     if I = FromCol then
      continue;
     //We have a filter so start filtering
     aFrm := fFilterArray.GetFilterForm(I);
     if aFrm.isFiltered then
     begin
      ValArry := aFrm.FilteredArray;

      for s in ValArry do
      begin
        if Piece(aStr, '^', I + 1) = s then
        begin
          fnd := true;
          break;
        end;
      end;

      if not fnd then
      begin
        result := false;
        exit;
      end;
     end;
    end;
  end;

var
  Idx: Integer;
begin
  if AddToFilter then
  begin
    Idx := fMasterList.IndexOfPiece(Value, U, ColumnNum + 1);
    while Idx > -1 do
    begin
      if CanAddWithFilters(fMasterList[Idx], ColumnNum) then
        fFilterList.Add(fMasterList[Idx]);
      Idx := fMasterList.IndexOfPieceEx(Value, U, ColumnNum + 1, Idx + 1);
    end;
  end
  else
  begin
    Idx := fFilterList.IndexOfPiece(Value, U, ColumnNum + 1);
    while Idx > -1 do
    begin
      fFilterList.Delete(Idx);
      Idx := fFilterList.IndexOfPiece(Value, U, ColumnNum + 1);
    end;
  end;
  lvResults.Items.Count := fFilterList.Count;
end;

procedure TfrmAudit.SetColumnWidth();
// ******************************************************************************
// Resize the virtual list views based on the new data
// ******************************************************************************

  procedure AutoResizeColumn(const AColumn: Integer);
  // ******************************************************************************
  // Auto resize a column
  // ******************************************************************************
  var
    s: string;
    I: Integer;
    MaxWidth: Integer;
    ItemWidth: Integer;
  begin
    // Grab the widht of the columns caption
    MaxWidth := ListView_GetStringWidth(lvResults.handle,
      PChar(lvResults.Columns.Items[AColumn].Caption));
    // Grab the width of the text that should display in that column
    for I := 0 to fFilterList.Count - 1 do
    begin
      s := Piece(fFilterList[I], U, AColumn + 1);
      if AColumn = 0 then
        s := Piece(s, ';', 1);

      if fFilterArray.FilterFrom[AColumn].FilterType = tFTFMDate then
        s := FormatDateTime('mm/dd/yyyy hh:nn:ss',
          FMDateTimeToDateTime(StrToFloatDef(s, -1)));
      // Get the text width
      ItemWidth := ListView_GetStringWidth(lvResults.handle, PChar(s));
      // Update our width if needed
      if MaxWidth < ItemWidth then
        MaxWidth := ItemWidth;
    end;
    // Pad the column
    MaxWidth := MaxWidth + 25;
    // Set the column width
    lvResults.Columns.Items[AColumn].Width := MaxWidth;
  end;

var
  I: Integer;
begin
  // Loop through all the columns in the virtual list view
  for I := 0 to lvResults.Columns.Count - 1 do
    AutoResizeColumn(I);
end;

{ TFilterArrayHelper }

function TFilterArrayHelper.AddFilter(const ColIdx: Integer): Integer;
// ********************************************************
// Add a new entry to the filter array
// ********************************************************
begin
  SetLength(self, Length(self) + 1);
  self[high(self)].ColumnIndex := ColIdx;
  self[high(self)].FilterForm := TFrmFilter.Create(nil);
  result := High(self);
end;

function TFilterArrayHelper.FindRecord(const ColIdx: Integer): Integer;
// ********************************************************
// Look up an array index based on the column
// ********************************************************
var
  I: Integer;
begin
  result := -1;
  for I := Low(self) to High(self) do
  begin
    if self[I].ColumnIndex = ColIdx then
    begin
      result := I;
      break;
    end;
  end;
end;

function TFilterArrayHelper.GetFilterForm(const ColIdx: Integer): TFrmFilter;
// ********************************************************
// Retrun the filter form based on the column
// ********************************************************
var
  tmpInt: Integer;
begin
  tmpInt := FindRecord(ColIdx);
  if tmpInt = -1 then
    tmpInt := AddFilter(ColIdx);
  result := self[tmpInt].FilterForm;
end;

procedure TFilterArrayHelper.ClearList;
// ********************************************************
// Clear out the filter array and free the forms
// ********************************************************
var
  I: Integer;
begin
  for I := Low(self) to High(self) do
  begin
    if assigned(self[I].FilterForm) then
      FreeAndNil(self[I].FilterForm);
  end;
  SetLength(self, 0);
end;

end.
