// Mandatory unit header per GUI SAC, Section 6.1.1
//	Package: Provider Role Tool GUI application
//	Date Created: August 2017
//  Site Name: CPRS EP1 Remote Development Team
//  Developers: Scott Stephenson (Delphi), Chris Bell (Delphi)
//             Robert Davis (M), Jamie Crumley (M)
//	Description: This frame displays while the application is connecting to
//               server
//	Note: This application requires OR_3_453 in order to run.

unit fConnectToServer;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ExtCtrls,
  VA508AccessibilityManager, VA508AccessibilityRouter, Vcl.Imaging.jpeg;

type
  TConnectionProgress = class(TFrame)
    pnlConnection: TPanel;
    pnlBottom: TPanel;
    pnlCopyright: TPanel;
    lblLegalCopyright: TMemo;
    pnl508Disclaimer: TPanel;
    lbl508Notice: TMemo;
    Panel3: TPanel;
    pnlLogo: TPanel;
    Image1: TImage;
    Panel2: TPanel;
    lblFileVersion: TStaticText;
    lblComments: TStaticText;
    lblInternalName: TStaticText;
    lblProductName: TStaticText;
    lblCompanyName: TStaticText;
    lblOriginalFileName: TStaticText;
    lblCRC: TStaticText;
    lblConnection: TStaticText;
    procedure TimerConnectionTimer(Sender: TObject);
  private
    TimerConnection: TTimer;
    procedure RegisterWithMGR(aCtrl: TWinControl; aMgr: TVA508AccessibilityManager);
  public
    function EstablishConnection: boolean;
    constructor Create(AOwner: TComponent; a508Manager: TVA508AccessibilityManager); reintroduce;
  end;

implementation

{$R *.dfm}

uses
  ORNet, ORfn, uUtilities, VAUtils, fReassignOrders;

const
  TX_NOT_AUTHORIZED_USER = 'You are not authorized to use Provider Role Tool.';
  TX_FAIL_TO_CONNECT =
    'Provider Role Tool could not connect to a server. ' +
    'Verify with your IT liason that the server is configured to support Provider Role Tool' ;


function TConnectionProgress.EstablishConnection: boolean;
// verify we can connect to server
// verify the user is authorized to run this app
const
  TX_MISMTCHC    = 'This is version %s of %s. The running server version is %s. The program cannot run until the client is updated.';
  TX_MISMTCHS    = 'This is version %s of %s. The running server version is %s. The program cannot run until the server is updated.';
  TC_VER         = 'Server/Client Incompatibility';
var
  sError:String;
  ClientVer, ServerVer: string;
  MisMchInt: Integer;
begin
  TimerConnection := TTimer.Create(self);
  try
  TimerConnection.Interval := 1000;
  TimerConnection.OnTimer := TimerConnectionTimer;
  TimerConnection.Enabled:= true;
  try
    lblConnection.Visible := true;
    // try to connect to the server
    result:= ConnectToServer(TX_OPTION);
    // ConnectToServer normally shows a failure message, so this is a bit redundant
    // But we want to make sure a connection failure is always associated with a message (or two)
    if not result then
      begin
        sError := '';
        if Assigned(RPCBrokerV) then
        begin
          if RPCBrokerV.RPCBError <> '' then
  //          sError := CRLF+CRLF+RPCBrokerV.RPCBError;
  //        InfoBox(TX_FAIL_TO_CONNECT+sError, 'Connection Failed', MB_ICONERROR);
            sError := RPCBrokerV.RPCBError + CRLF+CRLF;
            if (sError <> '') and (pos('Cancelled Login Process',sError)=0) then
              sError := sError + TX_FAIL_TO_CONNECT;
            if sError <> '' then
              InfoBox(sError, 'Connection Failed', MB_ICONERROR);
        end;
        exit;
      end;
    // verify the user is allowed to use this application (note use of short-circuit boolean)
    if result and (not IsAuthUser) then
    begin
      result:= false;
      InfoBox(TX_NOT_AUTHORIZED_USER, 'Error', MB_OK or MB_ICONERROR);
      exit;
    end;

    ClientVer := ClientVersion(Application.ExeName);
    ServerVer := ServerVersion(TX_OPTION, ClientVer);
    MisMchInt := CompareVersion(ServerVer, ClientVer);
    if (MisMchInt > 0) then
    begin
      InfoBox(Format(TX_MISMTCHC, [ClientVer,  ExtractFileName(Application.ExeName), ServerVer]), TC_VER, MB_OK);
      result := false;
    end else if (MisMchInt < 0) then
    begin
      InfoBox(Format(TX_MISMTCHS, [ClientVer,  ExtractFileName(Application.ExeName), ServerVer]), TC_VER, MB_OK);
      result := false;
    end;
  finally
    TimerConnection.enabled:= false;
    Self.Visible:= false;
  end;
  finally
    TimerConnection.Free;
  end;
end;

procedure TConnectionProgress.TimerConnectionTimer(Sender: TObject);
// this executes from formShow until server connection is complete
// just giving the user feedback that we're busy
var
  BusyIndicator: string;
begin
  Case TimerConnection.Tag of
    0:
    begin
      TimerConnection.Tag:= 1;
      BusyIndicator:= '.. '; // keep three chars to avoid visual jutter
    end;
    1:
    begin
      TimerConnection.Tag:= 2;
      BusyIndicator:= '...'; // keep three chars to avoid visual jutter
    end;
    else
    begin
      TimerConnection.Tag:= 0;
      BusyIndicator:= '.  '; // keep three chars to avoid visual jutter
    end;
  end;
  lblConnection.Caption:= format('Connecting to server%s', [BusyIndicator]);
end;

constructor TConnectionProgress.Create(AOwner: TComponent; a508Manager: TVA508AccessibilityManager);
var
 aMsg: String;
begin
 inherited Create(AOwner);
  lblCompanyName.Caption        := 'Developed by the ' + FileVersionValue(Application.ExeName, FILE_VER_COMPANYNAME);
  lblFileVersion.Caption        := FileVersionValue(Application.ExeName, FILE_VER_FILEVERSION);
  lblLegalCopyright.Text        := FileVersionValue(Application.ExeName, FILE_VER_LEGALCOPYRIGHT);
  lblOriginalFileName.Caption   := FileVersionValue(Application.ExeName, FILE_VER_ORIGINALFILENAME);  //patch
  lblProductName.Caption        := FileVersionValue(Application.ExeName, FILE_VER_PRODUCTNAME);
  lblComments.Caption           := FileVersionValue(Application.ExeName, FILE_VER_COMMENTS);  // version comment
  lbl508Notice.Text             := StringReplace(lbl508Notice.Text, #13#10, '', [rfReplaceAll]);
  lblInternalName.Caption       := FileVersionValue(Application.ExeName, FILE_VER_INTERNALNAME);
  lblCRC.Caption                := 'CRC: ' + IntToHex(CRCForFile(Application.ExeName), 8);
 if ScreenReaderSystemActive then
 begin
  RegisterWithMGR(self, a508Manager);
  if AOwner = FormReassignOrders then
  begin
    aMsg := lblProductName.Caption + ' ' + lblComments.Caption + ' ' + lbl508Notice.Text + ' ' + lblLegalCopyright.Text;
    GetScreenReader.Speak(aMsg);
  end;
 end;

end;

procedure TConnectionProgress.RegisterWithMGR(aCtrl: TWinControl; aMgr: TVA508AccessibilityManager);
var
  I: Integer;
begin
  if assigned(aMgr) then
  begin
    for I := 0 to aCtrl.ComponentCount - 1 do
    begin
      if aCtrl.Components[I] is TWinControl then
      begin
        aMgr.AccessData.EnsureItemExists(TWinControl(aCtrl.Components[I]));
        if aCtrl.Components[I].ComponentCount > 0 then
          RegisterWithMGR(TWinControl(aCtrl.Components[I]), aMgr)
      end;
    end;
  end;
end;

end.
