// Mandatory unit header per GUI SAC, Section 6.1.1
// Package: Provider Role Tool GUI application
// Date Created: Janurary 2020
// Site Name: CPRS EP1 Remote Development Team
// Developers: Chris Bell (Delphi) Jamie Crumley (M)
// Description: Audit screen
// Note: This application requires OR*3.0*453 in order to run.

unit fFilter;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants,
  System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.CheckLst, uUtilities,
  Vcl.ExtCtrls;

type
  // Filter types
  tFilterType = (tFTFMDate, tFTInteger, tFTString, tFTMultiString);

  TFrmFilter = class(TForm)
    ckFilter: TCheckListBox;
    chkAll: TCheckBox;
    btnOk: TButton;
    btnCancel: TButton;
    btnSortAsc: TButton;
    btnSortDesc: TButton;
    gp_top: TGridPanel;
    gp_bottom: TGridPanel;
    procedure chkAllClick(Sender: TObject);
    procedure ckFilterClickCheck(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnSortAscClick(Sender: TObject);
    procedure btnSortDescClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormShow(Sender: TObject);
  private
    { Private declarations }
    fReturn: tFilterRecArray;
    fFilterType: tFilterType;
    fMultiPiece: Char;
    fMulitNumber: Integer;
    fFullList: TStringList;
    Function CalculateFormWidth: Integer;
    function GetFilteredArray: tStrArray;
  public
    { Public declarations }
    procedure AddToFilters(const Value: string; const FilterType: tFilterType; CheckMarkIfExist: Boolean = false);
    procedure AddMultiPeiceToFilters(const Value: string;
      const PieceDelim: Char; Const PieceNum: Integer; CheckMarkIfExist: Boolean = false);
    Procedure Clear(ClearFullList: Boolean);
    function ShowForm(const aLeft, aTop: Integer): TModalResult;
    function IsVisible(aValue: String): Boolean;
    function IsFiltered: Boolean;
    property FilterResults: tFilterRecArray read fReturn;
    property FilterType: tFilterType read fFilterType;
    property FilteredArray: tStrArray read GetFilteredArray;
  end;

function CreateFilterForm(): TFrmFilter;

var
  FrmFilter: TFrmFilter;

implementation

uses
  orfn;

{$R *.dfm}

function CreateFilterForm(): TFrmFilter;
begin
  FrmFilter := TFrmFilter.Create(nil);
  result := FrmFilter;
end;

procedure TFrmFilter.AddToFilters(const Value: string;
  const FilterType: tFilterType; CheckMarkIfExist: Boolean = false);
// ********************************************************
// Add a string to the filter list
// ********************************************************
var
  rntInt: Integer;
begin
  rntInt := ckFilter.Items.IndexOf(Value);
  if rntInt = -1 then
  begin

    if not assigned(fFullList) then
      fFullList := TStringList.Create;

    fFullList.add(Value);
    rntInt := ckFilter.Items.add(Value);
    ckFilter.Checked[rntInt] := true;
    fFilterType := FilterType;

  end else begin
     If CheckMarkIfExist then
      ckFilter.Checked[rntInt] := true;
  end;
end;

procedure TFrmFilter.AddMultiPeiceToFilters(const Value: string;
  const PieceDelim: Char; Const PieceNum: Integer; CheckMarkIfExist: Boolean = false);
// ********************************************************
// Add a string to the filter list
// ********************************************************
var
  rntInt: Integer;
  addStr: String;
begin
  addStr := Piece(Value, PieceDelim, PieceNum);
  rntInt := ckFilter.Items.IndexOf(addStr);
  if rntInt = -1 then
  begin
    if not assigned(fFullList) then
      fFullList := TStringList.Create;
    fFullList.add(Value);
    fMultiPiece := PieceDelim;
    fMulitNumber := PieceNum;
    rntInt := ckFilter.Items.add(addStr);
    ckFilter.Checked[rntInt] := true;
    fFilterType := tFTMultiString;
  end else
  begin
    If CheckMarkIfExist then
      ckFilter.Checked[rntInt] := true;
  end;
end;

function StringListSortCompareDesc(List: TStringList;
  Index1, Index2: Integer): Integer;
// ********************************************************
// Will sort in reverse order based on filter type
// ********************************************************
begin
  result := AnsiCompareStr(List[Index2], List[Index1]);
end;

function StringListSortCompareAsc(List: TStringList;
  Index1, Index2: Integer): Integer;
// ********************************************************
// Will sort based on filter type
// ********************************************************
begin
  result := AnsiCompareStr(List[Index1], List[Index2]);
end;

function StringListSortCompareIntDesc(List: TStringList;
  Index1, Index2: Integer): Integer;
// ********************************************************
// Will sort in reverse order based on filter type
// ********************************************************
begin
  result := CompareInteger(StrToIntDef(List[Index2], 0),
    StrToIntDef(List[Index1], 0));
end;

function StringListSortCompareIntAsc(List: TStringList;
  Index1, Index2: Integer): Integer;
// ********************************************************
// Will sort based on filter type
// ********************************************************
begin
  result := CompareInteger(StrToIntDef(List[Index1], 0),
    StrToIntDef(List[Index2], 0));
end;

function StringListSortCompareFMDateDesc(List: TStringList;
  Index1, Index2: Integer): Integer;
// ********************************************************
// Will sort in reverse order based on filter type
// ********************************************************
begin
  result := CompareFloat(StrToFMDateTime(List[Index2]),
    StrToFMDateTime(List[Index1]));
end;

function StringListSortCompareFMDateAsc(List: TStringList;
  Index1, Index2: Integer): Integer;
// ********************************************************
// Will sort based on filter type
// ********************************************************
begin
  result := CompareFloat(StrToFMDateTime(List[Index1]),
    StrToFMDateTime(List[Index2]));
end;

procedure TFrmFilter.btnSortAscClick(Sender: TObject);
// ********************************************************
// Sort ascending
// ********************************************************
var
  SortList: TStringList;
  I: Integer;
  s: String;
begin
  // Use a string list to perform the sort
  SortList := TStringList.Create;
  try
    // Add the checkbox info to the stringlist
    for I := 0 to ckFilter.Count - 1 do
      SortList.add(ckFilter.Items[I] + '^' + BoolToStr(ckFilter.Checked[I]));

    // Custom sort
    case fFilterType of
      tFTFMDate:
        SortList.CustomSort(StringListSortCompareFMDateAsc);
      tFTInteger:
        SortList.CustomSort(StringListSortCompareIntAsc);
      tFTString:
        SortList.CustomSort(StringListSortCompareAsc);
      tFTMultiString:
        SortList.CustomSort(StringListSortCompareAsc);
    end;

    // Clear the display
    ckFilter.Clear;
    // Add items from the string list back to the display
    for s in SortList do
    begin
      I := ckFilter.Items.add(Piece(s, '^', 1));
      ckFilter.Checked[I] := StrToBoolDef(Piece(s, '^', 2), true);
    end;

    // Update the array
    for I := Low(fReturn) to High(fReturn) do
    begin
      if fFilterType = tFTMultiString then
       s := Piece(fReturn[I].Value, fMultiPiece, fMulitNumber)
      else if fFilterType = tFTFMDate then
        s := FormatFMDateTimeStr('mm/dd/yyyy hh:nn:ss', fReturn[I].Value)
      else
        s := fReturn[I].Value;
      fReturn[I].Index := ckFilter.Items.IndexOf(s);
    end;

  finally
    SortList.Free;
  end;
end;

procedure TFrmFilter.btnSortDescClick(Sender: TObject);
// ********************************************************
// Sort descending
// ********************************************************
var
  SortList: TStringList;
  I: Integer;
  s: String;
begin
  // Use a string list to perform the sort
  SortList := TStringList.Create;
  try
    // Add the checkbox info to the stringlist
    for I := 0 to ckFilter.Count - 1 do
      SortList.add(ckFilter.Items[I] + '^' + BoolToStr(ckFilter.Checked[I]));
    // Custom sort
    case fFilterType of
      tFTFMDate:
        SortList.CustomSort(StringListSortCompareFMDateDesc);
      tFTInteger:
        SortList.CustomSort(StringListSortCompareIntDesc);
      tFTString:
        SortList.CustomSort(StringListSortCompareDesc);
      tFTMultiString:
        SortList.CustomSort(StringListSortCompareDesc);
    end;

    // Clear the display
    ckFilter.Clear;
    // Add items from the string list back to the display
    for s in SortList do
    begin
      I := ckFilter.Items.add(Piece(s, '^', 1));
      ckFilter.Checked[I] := StrToBoolDef(Piece(s, '^', 2), true);
    end;

    // Update the array
    for I := Low(fReturn) to High(fReturn) do
    begin
      if fFilterType = tFTMultiString then
       s := Piece(fReturn[I].Value, fMultiPiece, fMulitNumber)
      else if fFilterType = tFTFMDate then
       s := FormatFMDateTimeStr('mm/dd/yyyy hh:nn:ss', fReturn[I].Value)
      else
       s := fReturn[I].Value;
      fReturn[I].Index := ckFilter.Items.IndexOf(s);
    end;
  finally
    SortList.Free;
  end;
end;

procedure TFrmFilter.chkAllClick(Sender: TObject);
// ********************************************************
// Check or uncheck all the entries
// ********************************************************
var
  I, aIdx: Integer;
  LookupStr: String;
begin
  // check or uncheck all the boxes
  if chkAll.Checked then
  begin
    // lookup the array record and if not exisitng then add
    for I := 0 to ckFilter.Items.Count - 1 do
    begin
      if fFilterType = tFTMultiString then
      begin
        aIdx := fFullList.IndexOfPiece(ckFilter.Items[I], fMultiPiece,
          fMulitNumber);
        LookupStr := fFullList[aIdx];
      end
      else if fFilterType = tFTFMDate then
        LookupStr := FloatToStr(StrToFMDateTime(ckFilter.Items[I]))
      else
        LookupStr := ckFilter.Items[I];

      aIdx := fReturn.FilterRecordIndex(I, true, LookupStr,
        ckFilter.Checked[I]);
      fReturn[aIdx].Action := tfaAdd;
    end;
    // Update the display
    ckFilter.CheckAll(cbChecked);

  end
  else
  begin
    // lookup the array record and if not exisitng then add
    for I := 0 to ckFilter.Items.Count - 1 do
    begin
      if fFilterType = tFTMultiString then
      begin
        aIdx := fFullList.IndexOfPiece(ckFilter.Items[I], fMultiPiece,
          fMulitNumber);
        LookupStr := fFullList[aIdx];
      end
      else if fFilterType = tFTFMDate then
        LookupStr := FloatToStr(StrToFMDateTime(ckFilter.Items[I]))
      else
        LookupStr := ckFilter.Items[I];

      aIdx := fReturn.FilterRecordIndex(I, true, LookupStr,
        ckFilter.Checked[I]);
      fReturn[aIdx].Action := tfaRemove;
    end;
    // Update the display
    ckFilter.CheckAll(cbUnchecked);
  end;

end;

procedure TFrmFilter.ckFilterClickCheck(Sender: TObject);
// ********************************************************
// Update the return array
// ********************************************************
var
  aIdx: Integer;
  LookupStr: String;
  rtnAct: TNotifyEvent;
begin
  // Lookup the array location and add if needed
  if fFilterType = tFTMultiString then
  begin
    aIdx := fFullList.IndexOfPiece(ckFilter.Items[ckFilter.ItemIndex],
      fMultiPiece, fMulitNumber);
    LookupStr := fFullList[aIdx];
  end
  else if fFilterType = tFTFMDate then
    LookupStr := FloatToStr(StrToFMDateTime(ckFilter.Items[ckFilter.ItemIndex]))
  else
    LookupStr := ckFilter.Items[ckFilter.ItemIndex];

  aIdx := fReturn.FilterRecordIndex(ckFilter.ItemIndex, true, LookupStr,
    not ckFilter.Checked[ckFilter.ItemIndex]);

  // Update the action
  if ckFilter.Checked[ckFilter.ItemIndex] then
    fReturn[aIdx].Action := tfaAdd
  else
    fReturn[aIdx].Action := tfaRemove;


  // Remove the onclick event to make the update
  rtnAct := chkAll.OnClick;
  chkAll.OnClick := nil;
  try
    chkAll.Checked := not IsFiltered;
  finally
    // Restore the onclick event
    chkAll.OnClick := rtnAct;
  end;


end;

procedure TFrmFilter.FormClose(Sender: TObject; var Action: TCloseAction);
// ********************************************************
// Handle the return of information
// ********************************************************
var
  I: Integer;
begin
  if ModalResult <> mrOk then
  begin
    // when canceling return the boxes to thier inital checks
    for I := Low(fReturn) to High(fReturn) do
      ckFilter.Checked[fReturn[I].Index] := fReturn[I].InitalChecked;

    // clear the return list since nothing changed
    fReturn.Clear;

  end
  else
  begin
    // If nothing really changed then remove the item from the return
    for I := high(fReturn) downto Low(fReturn) do
    begin
      if fReturn[I].InitalChecked = ckFilter.Checked[fReturn[I].Index] then
        fReturn.Delete(I);
    end;
  end;
end;

Function TFrmFilter.CalculateFormWidth: Integer;
var
  I, w, MaxWidth, tmp: Integer;
begin
  tmp := ckFilter.Width;
  MaxWidth := 0;
  for I := 0 to ckFilter.Items.Count - 1 do
    with ckFilter do
    begin
      w := Canvas.TextWidth(Items[I]);
      if w > MaxWidth then
        MaxWidth := w;
    end;
  result := self.Width + (MaxWidth + 40) - tmp;
end;

procedure TFrmFilter.FormCreate(Sender: TObject);
// ********************************************************
// Setup
// ********************************************************
begin
  fReturn.Clear;
end;

procedure TFrmFilter.FormDestroy(Sender: TObject);
// ********************************************************
// Tear Down
// ********************************************************
begin
  fReturn.Clear;
  FreeAndNil(fFullList);
end;

procedure TFrmFilter.FormShow(Sender: TObject);
// ********************************************************
// Screen show
// ********************************************************
var
  rtnAct: TNotifyEvent;
  I: Integer;
begin
  // Remove the onclick event to make the update
  rtnAct := chkAll.OnClick;
  chkAll.OnClick := nil;
  try
    chkAll.Checked := not IsFiltered;

    // If anything in the list then preset the initilization
    for I := Low(fReturn) to High(fReturn) do
      fReturn[I].InitalChecked := not ckFilter.Checked[fReturn[I].Index];

  finally
    // Restore the onclick event
    chkAll.OnClick := rtnAct;
  end;
end;

function TFrmFilter.GetFilteredArray: tStrArray;
var
  I, aIdx: integer;
  AddVal: String;
begin
  SetLength(Result, 0);
  for I := 0 to ckFilter.Count - 1 do
  begin
    if ckFilter.Checked[I] then
    begin
      if fFilterType = tFTMultiString then
      begin
        aIdx := fFullList.IndexOfPiece(ckFilter.Items[I], fMultiPiece,
          fMulitNumber);
        AddVal := fFullList[aIdx];
      end
      else if fFilterType = tFTFMDate then
        AddVal := FloatToStr(StrToFMDateTime(ckFilter.Items[I]))
      else
        AddVal := ckFilter.Items[I];

      SetLength(Result, Length(Result) + 1);
      Result[high(Result)] := AddVal;
    end;
  end;
end;

function TFrmFilter.ShowForm(const aLeft, aTop: Integer): TModalResult;
// ********************************************************
// Display the form
// ********************************************************
begin
  self.Left := aLeft;
  self.Top := aTop;
  self.Width := CalculateFormWidth;
  result := self.ShowModal;
end;

Procedure TFrmFilter.Clear(ClearFullList: Boolean);
var
  I: integer;
begin
 if ClearFullList then
 begin
  ckFilter.Clear;
  if assigned(fFullList)then
    fFullList.Clear;
 end else begin
  for I := 0 to ckFilter.Count - 1 do
    ckFilter.Checked[i] := false;

 end;


 fReturn.Clear;
end;

function TFrmFilter.IsFiltered: Boolean;
// ********************************************************
// Are items being filtered out
// ********************************************************
var
  I: Integer;
begin
  result := false;
  for I := 0 to ckFilter.Count - 1 do
  begin
    if not ckFilter.Checked[I] then
    begin
      result := true;
      break;
    end;
  end;
end;

function TFrmFilter.IsVisible(aValue: String): Boolean;
// ********************************************************
// Is the current item not filtered
// ********************************************************
var
  FndIdx: Integer;
begin
  result := false;
  FndIdx := ckFilter.Items.IndexOf(aValue);
  if FndIdx > -1 then
    result := ckFilter.Checked[FndIdx];

end;

end.
