// Mandatory unit header per GUI SAC, Section 6.1.1
// Package: Provider Role Tool GUI application
// Date Created: August 2017
// Site Name: CPRS EP1 Remote Development Team
// Developers: Scott Stephenson (Delphi), Chris Bell (Delphi)
// Robert Davis (M), Jamie Crumley (M)
// Description: Form that displays patient orders at left, shows detail at right
// Note: This application requires OR_3_453 in order to run.

unit fPatientOrders;

interface

uses
  uProviderEngine,
  fBase508Form,
  // standard Delphi units
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants,
  System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ExtCtrls, Vcl.ComCtrls, ORCtrls,
  Vcl.StdCtrls, VA508AccessibilityManager, System.Generics.Collections,
  System.Generics.Defaults, U508Extensions;

type
  TFormPatientOrders = class(TfrmBase508Form)
    pnlBottom: TPanel;
    clvOrders: TCaptionListView;
    Button1: TButton;
    spHorz: TSplitter;
    pnlRight: TPanel;
    pnlRBottom: TPanel;
    lblTransfer: TLabel;
    clvTranfers: TCaptionListView;
    Splitter1: TSplitter;
    MemoOrderDetail: TRichEdit;
    pnlInfo: TPanel;
    procedure clvOrdersSelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
    procedure clvOrdersDeletion(Sender: TObject; Item: TListItem);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    fGrpDict: TDictionary<String, TListGroup>;
    fLV508Manager: Tlv508Manager;
    procedure ShowOrders(AOrderEngine: tOrderEngine;
      APatientIEN: string); overload;
    procedure ShowOrders(AOrderEngine: tOrderEngine; aFromObj: TObject;
      const aFromIEN: String = ''); overload;
  end;

procedure ShowPatientOrders(AOrderEngine: tOrderEngine;
  APatientIEN, APatientName: string); overload;
procedure ShowPatientOrders(AOrderEngine: tOrderEngine; aFromObj: TObject;
  const aFromIEN: String = ''); overload;

implementation

{$R *.dfm}

uses
  uUtilities,
  VAUtils,
  uSizeEngine,
  fReassignOrders,
  fAudit;

procedure ShowPatientOrders(AOrderEngine: tOrderEngine;
  APatientIEN, APatientName: string);
var
  f: TFormPatientOrders;
begin
  f := TFormPatientOrders.Create(Application);
  try
    f.Caption := format('Qualifying Orders for Patient %s', [APatientName]);
    f.ShowOrders(AOrderEngine, APatientIEN);
    f.ShowModal;
  finally
    f.Release;
  end;
end;

procedure ShowPatientOrders(AOrderEngine: tOrderEngine; aFromObj: TObject;
  const aFromIEN: String = '');
var
  f: TFormPatientOrders;
begin
  f := TFormPatientOrders.Create(Application);
  try
    f.Caption := 'Qualifying Orders for Patient(s)';
    f.ShowOrders(AOrderEngine, aFromObj, aFromIEN);
    f.ShowModal;
  finally
    f.Release;
  end;
end;

procedure TFormPatientOrders.clvOrdersDeletion(Sender: TObject;
  Item: TListItem);
begin
  inherited;
  if assigned(Item.Data) then
    System.Dispose(pOrdRec(Item.Data));
end;

procedure TFormPatientOrders.clvOrdersSelectItem(Sender: TObject;
  Item: TListItem; Selected: Boolean);
var
  Xfers: TStringList;
  LstStr: String;
  aOrd: tOrderRecord;
begin
  if assigned(Item.Data) and Selected then
  begin
    aOrd := pOrdRec(Item.Data)^;
    MemoOrderDetail.Text := OrderDetails(aOrd.orderIEN, aOrd.patient.IEN);
    clvTranfers.Clear;
    // if aOrd.Xfer = xfer_PendOut then
    // begin
    Xfers := TStringList.Create;
    try
      GetTransferHistory(aOrd.orderIEN, Xfers);
      for LstStr in Xfers do
        clvTranfers.Add(LstStr);
    finally
      if assigned(Xfers) then
        Xfers.Free;
    end;
    // end;
  end;
end;

procedure TFormPatientOrders.FormCreate(Sender: TObject);
begin
  inherited;
  SizeHolder.LoadSize(self);
  SizeHolder.LoadSize(clvOrders);
  SizeHolder.LoadSize(pnlRBottom);
  SizeHolder.LoadSize(clvTranfers);
  fGrpDict := TDictionary<String, TListGroup>.Create(10009);
  if ScreenReaderActive then
  begin
    fLV508Manager := Tlv508Manager.Create;
    amgrMain.ComponentManager[clvOrders] := fLV508Manager;
    amgrMain.ComponentManager[clvTranfers] := fLV508Manager;
  end;
end;

procedure TFormPatientOrders.FormDestroy(Sender: TObject);
begin
  SizeHolder.SetSize(self);
  SizeHolder.SetSize(clvOrders);
  SizeHolder.SetSize(pnlRBottom);
  SizeHolder.SetSize(clvTranfers);

  FreeAndNil(fGrpDict);

  if assigned(fLV508Manager) then
  begin
    amgrMain.ComponentManager[clvOrders] := nil;
    amgrMain.ComponentManager[clvTranfers] := nil;
    FreeAndNil(fLV508Manager);
  end;

  inherited;
end;

procedure TFormPatientOrders.ShowOrders(AOrderEngine: tOrderEngine;
  APatientIEN: string);
var
  AOrder: tOrderRecord;
  AItem: TListItem;
  OldEvt: TLVChangeEvent;
begin
  OldEvt := clvOrders.OnChange;
  clvOrders.OnChange := nil;
  try
    clvOrders.Items.BeginUpdate;
    try
      for AOrder in AOrderEngine.Orders do
        if AOrder.patient.IEN = APatientIEN then
        begin
          AItem := clvOrders.Items.Add;
          AItem.Caption := AOrder.orderIEN;
          AItem.SubItems.Add(AOrder.OrderDate);
          AItem.SubItems.Add(AOrder.OrderStatus);
          // if AOrder.Xfer then
          // AItem.SubItems.Add('Transfer History');
        end;
    finally
      clvOrders.Items.EndUpdate;
    end;
  finally
    clvOrders.OnChange := OldEvt;
  end;
end;

procedure TFormPatientOrders.ShowOrders(AOrderEngine: tOrderEngine;
  aFromObj: TObject; const aFromIEN: String = '');

  procedure SetColumnWidth(awidth: Integer);
  var
    I: Integer;
  begin
    for I := 0 to clvOrders.Columns.Count - 1 do
      clvOrders.Columns[I].Width := awidth;
  end;

Const
  U = '^';
  ID_IEN = 1;
  ID_NAME = 2;

var
  AItem: TListItem;
  AOrders: tOrderRecords;
  AOrder: tOrderRecord;
  OldEvt: TLVChangeEvent;
  aPOrRec: pOrdRec;
  aListObj: pListObj;
  orderRec: tOrderRecord;
  I: Integer;
  nodeEnumerator: TTreeNodesEnumerator;
  ANode, ChildNode: TORTreeNode;
  XferRecord: tTransferToRec;
  FromAudit: Boolean;
  AddedOrders: TStringList;
  s: String;
begin
  OldEvt := clvOrders.OnChange;
  clvOrders.OnChange := nil;
  try
    clvOrders.Items.BeginUpdate;
    AddedOrders := TStringList.Create;
    try

      if aFromObj is TListView then
      begin
        // Loop through the items
        FormReassignOrders.ProgressStatusBar.ResetProgressBar
          (TListView(aFromObj).SelCount);
        FormReassignOrders.ProgressStatusBar.ResetLoop
          (TListView(aFromObj).SelCount);

        SetColumnWidth(0);

        for I := 0 to TListView(aFromObj).Items.Count - 1 do
        begin
          if FormReassignOrders.ProgressStatusBar.CancelClicked then
            break;
          if not TListView(aFromObj).Items[I].Selected then
            continue;

          FromAudit := false;
          Setlength(AOrders, 0);
          if (TListView(aFromObj) = FormReassignOrders.lvPendNone) or
            (TListView(aFromObj) = FormReassignOrders.lvPendOut) or
            (TListView(aFromObj) = FormReassignOrders.lvPendIn) then
          begin
            // Grab the order based on the selection
            if TListView(aFromObj) = FormReassignOrders.lvPendNone then
              orderRec := AOrderEngine.OrdersPendingNone
                [TListView(aFromObj).Items[I].Index]
            else if TListView(aFromObj) = FormReassignOrders.lvPendOut then
              orderRec := AOrderEngine.OrdersPendingOut
                [TListView(aFromObj).Items[I].Index]
            else if TListView(aFromObj) = FormReassignOrders.lvPendIn then
              orderRec := AOrderEngine.OrdersPendingIn
                [TListView(aFromObj).Items[I].Index];
            Setlength(AOrders, length(AOrders) + 1);
            AOrders[high(AOrders)] := orderRec;
          end
          else
          begin
            If assigned(frmAudit) then
            begin
              if TListView(aFromObj) = frmAudit.lvResults then
              begin
                if (AddedOrders.IndexOf(TListView(aFromObj).Items[I].SubItems[0]
                  ) = -1) then
                begin
                  Setlength(AOrders, length(AOrders) + 1);
                  AOrders[high(AOrders)].patient.name := TListView(aFromObj)
                    .Items[I].Caption;
                  s := frmAudit.AuditList[TListView(aFromObj).Items[I].Index];
                  AOrders[high(AOrders)].patient.IEN :=
                    Piece(Piece(s, U, 1), ';', 2);
                  AOrders[high(AOrders)].orderIEN := TListView(aFromObj)
                    .Items[I].SubItems[0];
                  AOrders[high(AOrders)].OrderDate := TListView(aFromObj)
                    .Items[I].SubItems[1];
                  AOrders[high(AOrders)].OrderStatus := TListView(aFromObj)
                    .Items[I].SubItems[2];
                  AOrders[high(AOrders)].orderXferEffective :=
                    TListView(aFromObj).Items[I].SubItems[3];
                  AddedOrders.Add(TListView(aFromObj).Items[I].SubItems[0]);
                end;
                FromAudit := true;
              end;
            end;

            if not FromAudit then
            begin
              if not assigned(TListView(aFromObj).Items[I].Data) then
                continue;
              aListObj := pListObj(TListView(aFromObj).Items[I].Data);
              orderRec := AOrderEngine.OrderByIEN(aListObj.orderIEN);
              case aListObj.Group_Type of
                // All by patient
                tvtPatient:
                  AOrders := AOrderEngine.ordersForPatient
                    (orderRec.patient.IEN);
                tvtXfer:
                  AOrders := AOrderEngine.ordersForPatientAndXfer
                    (orderRec.patient.IEN, orderRec.OrderXfer);
              end;
            end;
          end;

          for AOrder in AOrders do
          begin
            // create to order object to use later
            System.New(aPOrRec);
            aPOrRec.Assign(AOrder);
            AItem := clvOrders.Items.Add;
            AItem.Caption := AOrder.patient.name;
            AItem.SubItems.Add(AOrder.orderIEN);
            AItem.SubItems.Add(AOrder.OrderDate);
            AItem.SubItems.Add(AOrder.OrderStatus);
            if not FromAudit then
            begin
              case AOrder.OrderXfer of
                txftNone:
                  AItem.SubItems.Add('');
                txftTransferred:
                  AItem.SubItems.Add('Transferred');
                txftPendOut:
                  AItem.SubItems.Add('Pending Outgoing');
                txftPendIn:
                  AItem.SubItems.Add('Pending Incoming');
              end;
            end
            else
            begin
              if StrToDateTime(AOrder.orderXferEffective) >= now then
                AItem.SubItems.Add('Transferred')
              else
                AItem.SubItems.Add('Pending')
            end;

            AItem.Data := aPOrRec;

            if aFromIEN <> '' then
              if aFromIEN = AOrder.orderIEN then
                AItem.Selected := true;

          end;

          FormReassignOrders.ProgressStatusBar.StepProgress;
        end;
      end
      else if (aFromObj is TTreeView) then
      begin
        // Loop through the items
        FormReassignOrders.ProgressStatusBar.ResetProgressBar
          (TTreeView(aFromObj).SelectionCount);
        FormReassignOrders.ProgressStatusBar.ResetLoop
          (TTreeView(aFromObj).SelectionCount);

        SetColumnWidth(0);
        nodeEnumerator := TTreeNodesEnumerator.Create
          (TTreeView(aFromObj).Items);
        try
          while nodeEnumerator.moveNext do
          begin
            ANode := TORTreeNode(nodeEnumerator.Current);
            if (ANode.hasChildren) and (ANode.Selected) then
            begin
              // Select all the children
              ChildNode := TORTreeNode(ANode.getFirstChild);
              while assigned(ChildNode) do
              begin
                TTreeView(aFromObj).Select(ChildNode, [ssCtrl]);
                // TORTreeNode(childNode).Selected := True;
                ChildNode := TORTreeNode(ChildNode.GetNextSibling);
              end;
            end
            else if (ANode.level = 1) and (ANode.Selected) then
            begin
              XferRecord := FormReassignOrders.TransferOrders.
                TransferRec[ANode];

              case XferRecord.Group_Type of
                tvtPatient:
                  begin
                    AOrders := AOrderEngine.ordersForPatient
                      (XferRecord.PatientInfo.IEN);
                  end;
                tvtXfer:
                  begin
                    // aOldOrder := fOrderEngine.OrderByIEN(XferRecord.ListObject.orderIEN);
                    AOrders := AOrderEngine.ordersForPatientAndXfer
                      (XferRecord.PatientInfo.IEN,
                      XferRecord.PatientInfo.OrderXfer);
                  end;
                tvtOrder:
                  begin
                    Setlength(AOrders, 1);
                    AOrders[0] := AOrderEngine.OrderByIEN(XferRecord.orderIEN);
                  end;
              end;

              for AOrder in AOrders do
              begin
                // create to order object to use later
                System.New(aPOrRec);
                aPOrRec.Assign(AOrder);
                AItem := clvOrders.Items.Add;
                AItem.Caption := AOrder.patient.name;
                AItem.SubItems.Add(AOrder.orderIEN);
                AItem.SubItems.Add(AOrder.OrderDate);
                AItem.SubItems.Add(AOrder.OrderStatus);
                case AOrder.OrderXfer of
                  txftNone:
                    AItem.SubItems.Add('');
                  txftTransferred:
                    AItem.SubItems.Add('Transferred');
                  txftPendOut:
                    AItem.SubItems.Add('Pending Outgoing');
                  txftPendIn:
                    AItem.SubItems.Add('Pending Incoming');
                end;

                AItem.Data := aPOrRec;

                if aFromIEN <> '' then
                  if aFromIEN = AOrder.orderIEN then
                    AItem.Selected := true;

              end;

            end;
            FormReassignOrders.ProgressStatusBar.StepProgress;
          end;

        finally
          FreeAndNil(nodeEnumerator);
        end;

      end;
      FormReassignOrders.ProgressStatusBar.HideProgress;
      SetColumnWidth(-1);
    finally
      FreeAndNil(AddedOrders);
      clvOrders.Items.EndUpdate;
    end;
  finally
    clvOrders.OnChange := OldEvt;
  end;
end;

end.
