// Mandatory unit header per GUI SAC, Section 6.1.1
//	Package: Provider Role Tool GUI application
//	Date Created: August 2017
//  Site Name: CPRS EP1 Remote Development Team
// Developers: Scott Stephenson (Delphi), Chris Bell (Delphi)
//             Robert Davis (M), Jamie Crumley (M)
//	Description: Shows a dialog of all providers in system, with optional date fields
//	Note: This application requires OR_3_453 in order to run.

unit fSelectProvider;

// Programmer Notes...
// Returns true if modal result is mrOK and a valid provider is selected
// returns an application-defined provider record TProviderRecord
// date is displayed only for the current provider - setting both at once reduces calls to M server

interface

uses
  // application units
  uProviderEngine,
  uUtilities,
  fBase508Form,
  // VA component units
  ORFn,
  ORCtrls,
  ORDtTm,
  // Standard Delphi units
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants,
  System.Classes, Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs,
  Vcl.ExtCtrls, Vcl.StdCtrls,
  System.Actions, Vcl.ActnList, VA508AccessibilityManager, System.UITypes,
  U508Extensions;

type

  TFrmSelectProvider = class(TfrmBase508Form)
    cboPtProvider: TORComboBox;
    pnlBottom: TPanel;
    btnOK: TButton;
    btnCancel: TButton;
    MainActions: TActionList;
    ActionOK: TAction;
    ActionCancel: TAction;
    pnlDate: TPanel;
    lbl508HintDateTime: TVA508StaticText;
    pnlTop: TPanel;
    ordbStop: TORDateBox;
    LblStopDate: TLabel;
    ordbStart: TORDateBox;
    lblStartDate: TLabel;
    pnlTopMiddle: TPanel;
    chkIncTerm: TCheckBox;
    actShowTerm: TAction;
    procedure cboPtProviderNeedData(Sender: TObject; const StartFrom: string;
      Direction, InsertAt: Integer);
    procedure ActionOKExecute(Sender: TObject);
    procedure ActionCancelExecute(Sender: TObject);
    procedure MainActionsUpdate(Action: TBasicAction; var Handled: Boolean);
    constructor Create(AOwner: TComponent); override;
    procedure cboPtProviderDblClick(Sender: TObject);
    procedure cboPtProviderChange(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure ordbStartExit(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure ActionShowTermExecute(Sender: TObject);
  private
    fProvider: tPersonRecord;
    procedure GetProviderInformation;
    procedure SetGUIDates(AStartDate, AStopDate: TDate);
    procedure GetGUIDates(var AStartDate, AStopDate: TDate);
    procedure InitCombo;
    function DateRangeIsValid: boolean;
    function ValidText(aDateBox:TORDateBox): Boolean;
  public
  end;


  function SelectProviderWithDateRange(AProviderType: tProviderType; var AStartDate, AStopDate: TDate; var AProviderRecord: tPersonRecord): boolean;

implementation

{$R *.dfm}
uses
  uSizeEngine;

function SelectProviderWithDateRange(AProviderType: tProviderType; var AStartDate, AStopDate: TDate; var AProviderRecord: tPersonRecord): boolean;

 procedure SetFormConstraints(aFrm: TFrmSelectProvider);
 var
  ConBuff: Integer;
 begin
   ConBuff := aFrm.Width - aFrm.ClientWidth;
   if aFrm.pnlDate.Visible then
    // Label length + Label Margins + Panel Margins
    aFrm.Constraints.MinWidth := aFrm.lbl508HintDateTime.Width + (aFrm.lbl508HintDateTime.Margins.Left + aFrm.lbl508HintDateTime.Margins.Right) + (aFrm.pnlDate.Margins.Left + aFrm.pnlDate.Margins.Right) + ConBuff
   else
     aFrm.Constraints.MinWidth := (aFrm.btnCancel.BoundsRect.Right - aFrm.btnOK.BoundsRect.Left) + (aFrm.btnCancel.Margins.Right * 2) + ConBuff;
 end;

// the main form is asking us to show a list of providers, and return the selected provider record
var
  f: TFrmSelectProvider;
begin
  f:= TFrmSelectProvider.Create(Application);
  try
    f.pnlDate.Visible:= AProviderType = tptCurrent;
    SetFormConstraints(f);
    f.SetGUIDates(AStartDate, AStopDate); // show the current engine dates in controls, even if not visible
    f.fProvider:= AProviderRecord;
    f.InitCombo;

    case AProviderType of
      tptCurrent: f.Caption:= 'Select Current Provider and Order Dates';
      tptNew: f.Caption:= 'Add New Provider';
    end;
    f.ShowModal;

    if f.ModalResult = mrOK then
    begin
      AProviderRecord:=  f.fProvider;
      f.GetGUIDates(AStartDate, AStopDate);
    end;

    Result:= (F.ModalResult = mrOK) and AProviderRecord.Valid;
  finally
    f.Release;
  end;
end;


procedure TFrmSelectProvider.ActionCancelExecute(Sender: TObject);
// Close without doing anything
begin
  Close;
  ModalResult:= mrCancel;
end;


procedure TFrmSelectProvider.ActionOKExecute(Sender: TObject);
// if a provider is double-clicked, update fProvider, validate, and close
begin
  GetProviderInformation;
  if FProvider.Valid then
  begin
    Close;
    ModalResult:= mrOK;
  end;
end;


procedure TFrmSelectProvider.ActionShowTermExecute(Sender: TObject);
begin
  inherited;
  InitCombo;
end;

procedure TFrmSelectProvider.cboPtProviderChange(Sender: TObject);
// if selected item is changed, update fProvider
begin
 // GetProviderInformation;
end;


procedure TFrmSelectProvider.cboPtProviderDblClick(Sender: TObject);
// if a provider is double-clicked, it is the same as clicking the OK button
begin
  ActionOK.Execute;
end;


procedure TFrmSelectProvider.cboPtProviderNeedData(Sender: TObject; const StartFrom: string;
  Direction, InsertAt: Integer);
// the combo needs data, so call SubSetOfProviders to get it
var
  results: tStringList;
begin
  results:= tStringList.Create;
  try
    if actShowTerm.Checked then
      SubSetOfActiveAndInactivePersons(StartFrom, Direction, results)
    else
      SubSetOfProviders(StartFrom, Direction, results);
    cboPtProvider.ForDataUse(results);
  finally
    freeAndNil(results);
  end;
end;

constructor TFrmSelectProvider.Create(AOwner: TComponent);
// clear local variables at startup
begin
  inherited Create(AOwner);
  fProvider.Clear;
end;


function TFrmSelectProvider.ValidText(aDateBox:TORDateBox): Boolean;
var
  FMDT: TFMDateTime;
begin
  try
    FMDT := ServerParseFMDate(aDateBox.Text);
    aDateBox.DateSelected := FMDateTimeToDateTime(FMDT);
    Result := FMDateTimeToDateTime(FMDT) = aDateBox.DateSelected;
  except
    on E: Exception do
      begin
        MessageDLG(E.Message,mtError,[mbOK],0);
        Result := False;
      end;
  end;
end;

procedure TFrmSelectProvider.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  SizeHolder.SetSize(self);
  inherited;
end;

procedure TFrmSelectProvider.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  inherited;
  if pnlDate.Visible then
    begin // no need to check dates if they are not visible;
      if ModalResult = mrOK then
        begin
          CanClose := False;
          if not ValidText(ordbStart) then
            ordbStart.SetFocus
          else if not ValidText(ordbStop) then
            ordbStop.SetFocus
          else
            CanClose := True;
        end;
    end;
end;

procedure TFrmSelectProvider.FormCreate(Sender: TObject);
begin
  inherited;
  SizeHolder.LoadSize(self);
end;

procedure TFrmSelectProvider.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  inherited;
  if Key = VK_ESCAPE then
    ModalResult := mrCancel;
end;

procedure TFrmSelectProvider.FormResize(Sender: TObject);
// enforce some constraints so the combo box does not disappear on resizing
begin
  inherited;
  cboPtProvider.Constraints.MinHeight:= pnlBottom.Height * 2;
  pnlDate.Constraints.MinWidth:= ordbStop.Left + ordbStop.Width + 10;
end;


procedure TFrmSelectProvider.GetProviderInformation;
// grab the selected provider from the combo box and place in fProvider
begin
  if cboPtProvider.ItemIEN > 0 then
    FProvider.Initialize(cboPtProvider.Text, IntToStr(cboPtProvider.ItemIEN), True)
  else
    FProvider.Clear;
end;


procedure TFrmSelectProvider.MainActionsUpdate(Action: TBasicAction; var Handled: Boolean);
// dynamically update control state
begin
  ActionOK.Enabled:= (cboPtProvider.ItemIEN > 0) and DateRangeIsValid;
end;


procedure TFrmSelectProvider.ordbStartExit(Sender: TObject);
var
  FMDT: TFMDateTime;
  ordb: TORDateBox;
begin
  inherited;
  if ModalResult = mrCancel then
    exit;

  ordb := TORDateBox(Sender);
  try
    FMDT := ServerParseFMDate(ordb.Text);
    ordb.DateSelected := FMDateTimeToDateTime(FMDT);
  except
    on E: Exception do
      begin
        ShowMessage(E.Message);
        ordb.SetFocus;
      end;
  end;
end;

procedure TFrmSelectProvider.GetGUIDates(var AStartDate, AStopDate: TDate);
// configure engine to reflect the GUI date controls
begin
  if DateRangeIsValid then
  begin
    aStartDate:= ordbStart.dateSelected;
    aStopDate:= ordbStop.dateSelected;
  end;
end;

procedure TFrmSelectProvider.InitCombo;
begin
  cboPtProvider.Clear;
  cboPtProvider.InitLongList(Trim(piece(FProvider.name, '-', 1)));
  cboPtProvider.SelectByIEN(StrToInt64Def(FProvider.IEN, 0));
end;


function TFrmSelectProvider.DateRangeIsValid: boolean;
// verify the user has entered valid date range in GUI
begin
  result:= ordbStart.dateSelected <= ordbStop.dateSelected
end;


procedure TFrmSelectProvider.SetGUIDates(aStartDate, aStopDate: TDate);
// configure the GUI date controls to reflect current engine settings
begin
  // TODO: SDS -> set the date range so it cannot be later than the server date
  ordbStart.dateSelected:= aStartDate;
  ordbStop.dateSelected:= aStopDate;
  ordbStart.FMDateTime := DateTimeToFMDateTime(aStartDate);
  ordbStop.FMDateTime:= DateTimeToFMDateTime(aStopDate);
end;


end.
