// Mandatory unit header per GUI SAC, Section 6.1.1
// Package: Provider Role Tool GUI application
// Date Created: August 2017
// Site Name: CPRS EP1 Remote Development Team
// Developers: Scott Stephenson (Delphi) Robert Davis (M)
// Description: This is the engine unit that supports order transfers
// Note: This application requires OR_3_453 in order to run.

unit uProviderEngine;

// non-GUI support engine and declarations for Provider Role Tool

// TOrderEngine can...
// - obtain a list of qualifying orders for a given provider & date range
// - reassign those orders to another provider
// - maintain an activity log of reassignment activity for later review

// M Calls...
// - ORQ3 EN -> returns list of patient/order records for a given provider date range
// - ORQ3 XFER -> reassign one or more orders to a new provider

interface

uses
  classes, types, Vcl.ComCtrls, System.Generics.Collections,
  System.Generics.Defaults, uUtilities;

type
  tOrderEngine = class
  private
    fCurrentProvider: tPersonRecord;
    fDateStart: tDate;
    fDateStop: tDate;
    fOrders: tOrderRecords;
    fPendInArray: tOrderRecords;
    fPendNoneArray: tOrderRecords;
    fPendOutArray: tOrderRecords;
    fWorking: Boolean;
    fSortType: tSortType;
    fTotalPatients: integer;
    fTotalXferPatIn: integer;
    fTotalXferPatOut: integer;
    fTotalXferPatNone: integer;
    fPatDict: TPersonDict;
    procedure ClearOrders;
    procedure AddToPatDict(aValue: String);
  public
    constructor Create;
    destructor Destroy; override;
    procedure getProviderOrdersFromServer(aPSB: tProgressStatusBar);
    function ordersForPatient(aPatientIEN: string): tOrderRecords;
    Procedure SetupOrdersBySeperateXfer; overload;
    Procedure SetupOrdersBySeperateXfer(aOrder: tOrderRecord); overload;
    function ordersForPatientAndXfer(const aPatientIEN: string;
      const aXfer: tXfertType): tOrderRecords;
    Function OrderByIEN(const aIEN: String): tOrderRecord;
    Function IndexByIen(const aIEN: String): Integer;
    function reassignOrders(var newAssignments: tReassignOrderRecords): integer;
    function reassignOrdersSimulated(var newAssignments: tReassignOrderRecords;
      succeed: Boolean): Boolean;
    procedure SortByIEN;
    procedure SortByOrderIEN;
    Procedure SortByXfer;
    procedure UpdateOrderFlagByPatient(const Flag: tOrderFlag; Const PatientIEN: String);
    procedure UpdateOrderFlagByPatientAndXfer(const Flag: tOrderFlag; Const PatientIEN: String; const aXfer: tXfertType);
    procedure UpdateOrderFlagByOrder(const Flag: tOrderFlag; Const OrderIEN: String);
    procedure ClearOrderFlagsByPatient(Const PatientIEN: String);
    procedure ClearOrderFlagsByPatientAndXfer(Const PatientIEN: String; const aXfer: tXfertType);
    procedure ClearOrderFlagsByOrder(Const OrderIEN: String);
    // property setters/getters
    procedure setDateStart(aDate: tDate);
    procedure setDateStop(aDate: tDate);
    // properties
    property CurrentProvider: tPersonRecord read fCurrentProvider
      write fCurrentProvider;
    property DateStart: tDate read fDateStart write setDateStart;
    property DateStop: tDate read fDateStop write setDateStop;
    //Holds all orders and cant be adjusted
    property Orders: tOrderRecords read fOrders;
    //Holds what is displayed and will varry
    property OrdersPendingOut: tOrderRecords read fPendOutArray;
    property OrdersPendingIn: tOrderRecords read fPendInArray;
    property OrdersPendingNone: tOrderRecords read fPendNoneArray;

    property Working: Boolean read fWorking;
    property TotalPatients: integer read fTotalPatients;
    property TotalPatientsXferIn: integer read fTotalXferPatIn;
    property TotalPatientsXferOut: integer read fTotalXferPatOut;
    property TotalPatientsXferNone: integer read fTotalXferPatNone;
  end;

implementation

uses
  SysUtils,
  DateUtils,
  orNet,
  orFn,
  fReassignOrders;

constructor tOrderEngine.Create;
begin
  inherited Create;
  fWorking := false;
  fCurrentProvider.Clear;
  fDateStop := Date;
  // we don't use FMToday because we may not be connected to server
  fDateStart := IncYear(fDateStop, -1);
  fPatDict := TDictionary<String, String>.Create(10009);
  fTotalPatients := 0;
  fTotalXferPatIn := 0;
  fTotalXferPatOut := 0;
  fTotalXferPatNone := 0;
end;

destructor tOrderEngine.Destroy;
begin
  FreeAndNil(fPatDict);
  Inherited;
end;

function tOrderEngine.ordersForPatient(aPatientIEN: string): tOrderRecords;
// return an array of order records where the patientIEN matches
var
  idx: integer;
  aOrder: tOrderRecord;
begin
  idx := 0;
  setLength(result, length(fOrders));
  for aOrder in fOrders do
    if aOrder.patient.IEN = aPatientIEN then
    begin
      result[idx] := aOrder;
      inc(idx);
    end;
  setLength(result, idx);
end;

Procedure tOrderEngine.SetupOrdersBySeperateXfer;
var
 order: tOrderRecord;
begin
  SortByXfer;

    setLength(fPendNoneArray, 0);
    setLength(fPendInArray, 0);
    setLength(fPendOutArray, 0);

    for order in Orders do
      SetupOrdersBySeperateXfer(order);
end;

Procedure tOrderEngine.SetupOrdersBySeperateXfer(aOrder: tOrderRecord);
begin
  case aOrder.OrderXfer of
    txftPendOut:
      fPendOutArray.Add(aOrder);
    txftPendIn:
      fPendInArray.Add(aOrder);
  else
    fPendNoneArray.Add(aOrder);
  end;
end;

function tOrderEngine.ordersForPatientAndXfer(const aPatientIEN: string;
  const aXfer: tXfertType): tOrderRecords;
// return an array of order records where the patientIEN and xfer type matches
var
  idx: integer;
  aOrder: tOrderRecord;
begin
  idx := 0;
  setLength(result, length(fOrders));
  for aOrder in fOrders do
    if (aOrder.patient.IEN = aPatientIEN) and (aOrder.OrderXfer = aXfer) then
    begin
      result[idx] := aOrder;
      inc(idx);
    end;
  setLength(result, idx);
end;

procedure tOrderEngine.getProviderOrdersFromServer(aPSB: tProgressStatusBar);
// calls ORQE3 EN to get a list of appropriate orders for provider X
var
  AStart: double;
  AStop: double;
  i: integer;
  results: TStringList;
begin
  fOrders := nil;
  fPatDict.Clear;
  fTotalPatients := 0;
  fTotalXferPatIn := 0;
  fTotalXferPatOut := 0;
  fTotalXferPatNone := 0;
  fWorking := true;
  try
    results := TStringList.Create;
    AStart := DateTimeToFMDateTime(fDateStart);
    AStop := DateTimeToFMDateTime(fDateStop);
    CallVistA('ORQ3 EN', [fCurrentProvider.IEN, AStart, AStop], results);
    // results.LoadFromFile('test.txt');

    setLength(fOrders, results.Count);
    aPSB.ResetLoop(results.Count);
    aPSB.ResetProgressBar(results.Count * 2);
    for i := 0 to results.Count - 1 do
    begin
      fOrders[i].LoadFromString(results[i]);
      AddToPatDict(results[i]);
      aPSB.StepProgress;
    end;
    fSortType := tstNone;
  finally
    fWorking := false;
    FreeAndNil(results);
  end;
end;

procedure tOrderEngine.AddToPatDict(aValue: String);
var
  PatIEN, RtnDummy: String;
  tmpxfer: integer;
begin
  PatIEN := piece(aValue, '^', 2);
  RtnDummy := '';
  if not fPatDict.TryGetValue(PatIEN, RtnDummy) then
  begin
    inc(fTotalPatients);
    fPatDict.Add(PatIEN, PatIEN);
    tmpxfer := StrToIntDef(piece(aValue, '^', 6), 0);
    case tmpxfer of
      0:
        inc(fTotalXferPatNone);
      1:
        inc(fTotalXferPatOut);
      2:
        inc(fTotalXferPatIn);
    end;

  end;
end;

function tOrderEngine.reassignOrders(var newAssignments
  : tReassignOrderRecords): integer;
// this function returns 0 for total success, else it returns a count of failed assignments
// EffectiveDateTime is DELPHI date time - we convert here to FMDateTime
// calls ORQ3 XFER [stringList] to reassign orders from old to new providers
// input format: OrderIEN^OldProvIEN^NewProvIEN^UserIEN^TransferDateTime in FileMan format
// ORQ3 XFER returns a string list of results by order
// output format: OrderIEN^ResultCode^errorMessage  -> result code 0 is fail, 1 is pass for each record
var
  i: integer;
  s: string;
  OrderList: TStringList;
  ResultsList: TStringList;
  OneOrder: tReassignOrderRecord;
  ActivityString: string;

  procedure UpdateOrderResult(s: string);
  // really crude means of updating each record's results
  // not optimized for speed, but should be OK with the record counts we expect (hundreds to thousands at most)
  var
    idx: integer;
    AOrderIEN: string;
  begin
    AOrderIEN := piece(s, U, 1);
    for idx := 0 to length(newAssignments) - 1 do
      if newAssignments[idx].orderIEN = AOrderIEN then
      begin
        newAssignments[idx].result := piece(s, U, 2);
        newAssignments[idx].errorMessage := piece(s, U, 3);
        break;
      end;

  end;

begin
  result := 0;
  fWorking := true;
  // create a list of changes, each is a ^ delimited string
  OrderList := TStringList.Create;
  ResultsList := TStringList.Create;
  try
    for OneOrder in newAssignments do
    begin
      s := format('%s^%s^%s^%s^%s^%s', [OneOrder.orderIEN, OneOrder.oldProviderIEN,
        OneOrder.newProviderIEN, OneOrder.requesterIEN,
        OneOrder.effectiveDateTime, OneOrder.orderFlag]);
      OrderList.Add(s);
    end;

    // perform the server command to reassign orders
    if OrderList.Count > 0 then
    begin
      // make the actual call to the server
      CallVistA('ORQ3 XFER', [OrderList], ResultsList);
    end;

    // TODO: create a user-friendly transfer log
    ActivityString := 'Reassigned Orders As Follows...' + sLineBreak;
    for s in ResultsList do
    begin
      ActivityString := ActivityString + s + sLineBreak;
    end;

    // identify any failed order transfers to the GUI by setting TransferResult in NewAssignments
    for i := 0 to ResultsList.Count - 1 do
    begin
      s := ResultsList[i];
      UpdateOrderResult(s);
      if piece(s, U, 2) = '0' then // 0 indicates failure
        inc(result);
    end;

  finally
    fWorking := false;
    FreeAndNil(OrderList);
    FreeAndNil(ResultsList);
  end;

end;

function tOrderEngine.reassignOrdersSimulated(var newAssignments
  : tReassignOrderRecords; succeed: Boolean): Boolean;
// this is a debug only method
var
  i: integer;
begin
  result := succeed;
  for i := 0 to length(newAssignments) - 1 do
  begin
    if succeed then
    begin
      newAssignments[i].result := '1';
      newAssignments[i].errorMessage := '';
    end
    else
    begin
      newAssignments[i].result := '0';
      newAssignments[i].errorMessage := 'Simulated error message';
    end;
  end;
end;

procedure tOrderEngine.SortByIEN;
begin
  if fSortType <> tstIEN then
  begin
    TArray.Sort<tOrderRecord>(fOrders,
      TDelegatedComparer<tOrderRecord>.Construct(
      function(const Left, Right: tOrderRecord): integer
      begin
        result := TComparer<Integer>.Default.Compare(StrToIntDef(Left.patient.IEN, -1),
          StrToIntDef(Right.patient.IEN, -1));
      end));
    fSortType := tstIEN;
  end;
end;

procedure tOrderEngine.SortByOrderIEN;
begin
  if fSortType <> tstOrderIEN then
  begin
    TArray.Sort<tOrderRecord>(fOrders,
      TDelegatedComparer<tOrderRecord>.Construct(
      function(const Left, Right: tOrderRecord): integer
      begin
        result := TComparer<Integer>.Default.Compare(StrToIntDef(Left.orderIEN, -1),
          StrToIntDef(Right.orderIEN, -1));
      end));
    fSortType := tstOrderIEN;
  end;
end;

Procedure tOrderEngine.SortByXfer;
begin
  if fSortType <> tstXfer then
  begin
    TArray.Sort<tOrderRecord>(fOrders,
      TDelegatedComparer<tOrderRecord>.Construct(
      function(const Left, Right: tOrderRecord): integer
      begin
        if Left.OrderXfer = Right.OrderXfer then
          result := TComparer<String>.Default.Compare(Left.patient.name,
            Right.patient.name)
        else
          result := TComparer<tXfertType>.Default.Compare(Left.OrderXfer,
            Right.OrderXfer);
      end));

    fSortType := tstXfer;
  end;
end;

Function tOrderEngine.IndexByIen(const aIEN: String): Integer;
// ********************************************************
// Return the index in the orders array for the order by IEN
// ********************************************************
var
  aOrder: tOrderRecord;
begin
  Result := -1;

  //Sort by IEN to speed up the search
  SortByOrderIEN;

  //Perform the search
  aOrder.orderIEN := aIEN;
  TArray.BinarySearch<tOrderRecord>(fOrders, aOrder, Result,
    TDelegatedComparer<tOrderRecord>.Construct(
    function(const Left, Right: tOrderRecord): Integer
    begin
      if StrToIntDef(Left.orderIEN, -1) > StrToIntDef(Right.orderIEN, -1) then Exit(1);
      if StrToIntDef(Left.orderIEN, -1) < StrToIntDef(Right.orderIEN, -1) then Exit(-1);
      Result:= 0;
    end
  ));
end;

Function tOrderEngine.OrderByIEN(const aIEN: String): tOrderRecord;
// ********************************************************
// Return the order record by order IEN
// ********************************************************
var
  fndIdx: Integer;
begin
  fndIdx := IndexByIen(aIEN);
  if fndIdx <> -1 then
    result := fOrders[FndIdx];
end;

procedure tOrderEngine.UpdateOrderFlagByPatient(const Flag: tOrderFlag; Const PatientIEN: String);
var
  I: Integer;
begin
  for I := Low(fOrders) to High(fOrders) do
  begin
     if (fOrders[i].patient.IEN = PatientIEN) then
       fOrders[i].orderFlag := flag;
  end;

end;

procedure tOrderEngine.UpdateOrderFlagByPatientAndXfer(const Flag: tOrderFlag; Const PatientIEN: String; const aXfer: tXfertType);
var
  I: Integer;
begin
  for I := Low(fOrders) to High(fOrders) do
  begin
     if (fOrders[i].patient.IEN = PatientIEN) and (fOrders[i].OrderXfer = aXfer) then
       fOrders[i].orderFlag := flag;
  end;
end;

procedure tOrderEngine.UpdateOrderFlagByOrder(const Flag: tOrderFlag; Const OrderIEN: String);
var
  fndIdx: Integer;
begin
  fndIdx := IndexByIen(OrderIEN);
  if fndIdx <> -1 then
    fOrders[FndIdx].orderFlag := Flag;
end;

procedure tOrderEngine.ClearOrderFlagsByPatient(Const PatientIEN: String);
begin
  UpdateOrderFlagByPatient(tofNA, PatientIEN);
end;

procedure tOrderEngine.ClearOrderFlagsByOrder(Const OrderIEN: String);
var
  fndIdx: Integer;
begin
  fndIdx := IndexByIen(OrderIEN);
  if fndIdx <> -1 then
    fOrders[FndIdx].orderFlag := tofNA;
end;

procedure tOrderEngine.ClearOrderFlagsByPatientAndXfer(Const PatientIEN: String; const aXfer: tXfertType);
var
  I: integer;
begin
  for I := low(fOrders) to high(fOrders) do
  begin
    if (fOrders[i].patient.IEN = PatientIEN) and (fOrders[i].OrderXfer = aXfer) and (fOrders[i].orderFlag <> tofNA) then
      fOrders[i].orderFlag := tofNA;
  end;
end;

procedure tOrderEngine.setDateStart(aDate: tDate);
// ********************************************************
// property setter for DateStart
// ********************************************************
begin
  if aDate <> fDateStart then
  begin
    fDateStart := aDate;
    ClearOrders;
  end;
end;

procedure tOrderEngine.setDateStop(aDate: tDate);
// ********************************************************
// property setter for DateStop
// ********************************************************
begin
  if aDate <> fDateStop then
  begin
    fDateStop := aDate;
    ClearOrders;
  end;
end;

procedure tOrderEngine.ClearOrders;
// ********************************************************
// since an order parameter has changed, internal data
// should be erased for consistency
// ********************************************************
begin
  fOrders := nil;
end;


end.
