// Mandatory unit header per GUI SAC, Section 6.1.1
// Package: Provider Role Tool GUI application
// Date Created: August 2017
// Site Name: CPRS EP1 Remote Development Team
// Developers: Scott Stephenson (Delphi) Robert Davis (M)
// Description: RPC and other nonvisual support utilites
// Note: This application requires OR*3.0*453 in order to run.

unit uUtilities;

interface

{
  M calls made by provider role tool application
  Many of these Delphi wrappers were simply taken as-is from CPRS

  ORWU DT [NOW] ->  to get server time
  ORQOR DETAIL [ORDER IEN, PATIENT IEN]-> to get order details
  ORWU NEWPERS [PROVIDER] -> to get a list of providers
  ORQ3 EN [PROVIDER IEN, START DATE, END DATE] -> gets a list of provider orders
  ORQ3 XFER [ORDER LIST] -> transfers orders between providers
  ORWU VERSRV [] -> use by ORNet unit to connect to Vista server and authorize M call subset
  ORQ3 AUTHUSR [] -> is this user authorized to use the application (RPC checks the "OR PRT ACCESS" key)
  ORWU USERINFO [] -> gets user information
  ORQ3 AUDIT --> Returns the list of data within date range
}

uses
  // Delphi units...
  SysUtils,
  Classes,
  DateUtils,
  messages,
  Comctrls,
  System.Generics.Collections,
  System.Generics.Defaults,
  Winapi.Windows,
  Forms,
  system.uitypes,
  vcl.Dialogs,
  Vcl.StdCtrls,
   // Application units...
//  uProg,
  orNet,
  orFn,
  U508Extensions;

type
  tProviderType = (tptCurrent, tptNew);
  tXfertType    = (txftNone, txftPendOut, txftPendIn, txftTransferred);
  tSortType     = (tstNone, tstIEN, tstXfer, tstOrderIEN);
  tViewType     = (tvtPatient, tvtXfer, tvtOrder);
  tXferReason   = (txrNone, txrSame, txrPend, txrCancel);
  TFilterAction = (tfaAdd, tfaRemove, tfaNone);
  tOrderFlag    = (tofNA, tofCancel, tofOverwrite);
  tTaskDLGType  = (tdlgNone, tdlgWarning, tdlgError, tdlgInformation, tdlgShield);

  pOrdRec = ^tOrderRecord;
  pListObj = ^tListObject;

  tSrchDictGroup = TDictionary<String, TListGroup>;
  tSrchDict = TDictionary<String, TListItem>;
  TPersonDict = TDictionary<String, String>;
  tStrArray = array of string;
  tDropLocations = Array of TListView;

  TUserInfo = record // record for ORWU USERINFO - stolen from CPRS
    DUZ: Int64;
    Name: string;
    UserClass: Integer;
    CanSignOrders: Boolean;
    IsProvider: Boolean;
    OrderRole: Integer;
    NoOrdering: Boolean;
    DTIME: Integer;
    CountDown: Integer;
    EnableVerify: Boolean;
    NotifyAppsWM: Boolean;
    PtMsgHang: Integer;
    Domain: string;
    Service: Integer;
    AutoSave: Integer;
    InitialTab: Integer;
    UseLastTab: Boolean;
    EnableActOneStep: Boolean;
    WebAccess: Boolean;
    IsRPL: string;
    RPLList: string;
    HasCorTabs: Boolean;
    HasRptTab: Boolean;
    IsReportsOnly: Boolean;
    ToolsRptEdit: Boolean;
    DisableHold: Boolean;
    GECStatusCheck: Boolean;
    StationNumber: string;
    IsProductionAccount: Boolean;
  end;

  tPersonRecord = record
    IEN: string;
    name: string;
    tag: integer; // multi-use, for example, store the patient's order count
    IsProvider: Boolean;
    OrderXfer: tXfertType; // Only used when orders are gouped by xfer type
    Procedure Assign(const aSrc: tPersonRecord);
    function ToStringData: string;
    function ToNodeName: string;
    procedure Clear;
    function Valid: Boolean;
    procedure Initialize(aName: string; aIEN: string; aIsProivder: Boolean);
  end;

  tOrderRecord = record
    // here we store an individual result returned by the M call "ORQ3 EN"
    patient: tPersonRecord;
    orderIEN: string;
    orderStatus: string;
    orderDate: string;
    OrderXfer: tXfertType;
    orderXferEffective: string; // Transfer effective date
    orderXferFrom: tPersonRecord; // Transferred from provider
    orderFlag: tOrderFlag;
    Procedure Assign(const aSrc: tOrderRecord);
    function CompareRecord(const aCompare: tOrderRecord): Boolean;
    Procedure LoadFromString(const aValue: String);
  end;

  tListObject = Record
    Group_Type: tViewType;
    PatientIEN: String;
    orderIEN: String;
    NumOfOrders: integer;
    Procedure Assign(const aSrc: tListObject);
  end;

  tTransferToRec = Record
    NodePointer: Integer;
    aTreeNode: TTreeNode;
    TransferTo: tPersonRecord;
    PatientInfo: tPersonRecord;
    OrderIEN: String;
    Group_Type: tViewType;
    NumOfOrders: integer;
    IsProviderDisplay: Boolean;
    OrderFlag: tOrderFlag;
    Procedure Assign(const aSrc: tTransferToRec);
  end;

  tReassignOrderRecord = record
    // here we store a record to passed to the M call "ORQ3 XFER"
    effectiveDateTime: string;
    requesterIEN: string;
    orderIEN: string;
    orderFlag: String;
    PatientIEN: string;
    patientName: string;
    oldProviderIEN: string;
    oldProviderName: string;
    newProviderIEN: string;
    newProviderName: string;
    result: string;
    errorMessage: string;
  end;

  tPatRec = record
    SelObj: TObject;
    name: String;
    IEN: String;
    Reason: tXferReason;
  end;

  tFilterRec = Record
    Index: Integer;
    Value: String;
    Action: TFilterAction;
    InitalChecked: Boolean;
  end;

  THelperListGroup = class helper for TListGroup
  public
    procedure IncFooterCount(PieceNum: Integer);
  end;

  THelperListGroups = class helper for TListGroups
  public
    function FindItemByGroupID(GroupID: Integer): TListGroup;
    function FindItemByGroupHeader(GroupHeader: String): TListGroup;
    function FindLastGroupID: Integer;
    function FindItemByGroup(aLookup: String; const SrchDict: tSrchDictGroup)
      : TListGroup;
  end;

  THelperListView = class helper for TListView
  public
    function FindListItemByPatient(const aIEN: string;
      const SrchDict: tSrchDict): TListItem;
    function FindListItemByXfer(const aIEN: string; const XferType: tXfertType;
      const SrchDict: tSrchDict): TListItem;
    Function AddListItemGroupByPatient(const aOrder: tOrderRecord;
      const SrchDict: tSrchDict): TListItem;
    Function AddListItemGroupByName(const aOrder: tOrderRecord;
      const SrchDict: tSrchDictGroup; CreatsGroups: Boolean): TListItem;
    Function AddListItemGroupByXfer(const aOrder: tOrderRecord;
      const SrchDict: tSrchDict): TListItem;
    Function CleanUpListItem(const aItem: TListItem): Boolean;
    Procedure IncOrderCount(const aItem: TListItem; IncBy: Integer = 1);
  end;

  tStringlistHelper = class helper for TStringList
    function IndexOfPiece(const S: string; const PieceDelim: Char; const PieceNum: integer): Integer;
    function IndexOfPieceEx(const S: string; const PieceDelim: Char; const PieceNum: integer; const StartFrom: Integer): Integer;
    function FindByPiece(const S: string; var Index: Integer; const PieceDelim: Char; const PieceNum: integer): Boolean;
    function FindByPieceEx(const S: string; var Index: Integer; const PieceDelim: Char; const PieceNum: integer;const StartFrom: Integer): Boolean;
  end;

  tOrderRecords = array of tOrderRecord;
  tPersonRecords = array of tPersonRecord;
  tReassignOrderRecords = array of tReassignOrderRecord;
  tPatRecArry = Array of tPatRec;
  tTransferOders = Array of tTransferToRec;
  tAnsiStringArray = array of AnsiString;
  tTreeNodeArray = array of TTreeNode;
  tFilterRecArray = array of tFilterRec;

  ttFilterRecHelper = record helper for tFilterRecArray
    private
      function GetFilterRecIndex(const aIdx: Integer; const AutoAdd: Boolean = false; const aValue: String = ''; const aChecked: Boolean = false): Integer;
      function GetFilterRec(const aIdx: Integer): tFilterRec;
    //  procedure DeleteIdx(const Index: Cardinal);
    public
      function FilterRecordIndex(const aIdx: Integer; const AutoAdd: Boolean = false; const aValue: String = ''; const aChecked: Boolean = false): Integer;
      property FilterRec[const aIdx: Integer]: tFilterRec read GetFilterRec;
      procedure Clear;
      function Delete(const aIdx: Integer): Boolean;
  end;

  TTransferOdersHelper = record helper for tTransferOders
    private
      function GetTransferRec(const aNode: TTreeNode): tTransferToRec;
     // procedure DeleteIdx(const Index: Cardinal);
      function FindRecord(const aNode: TTreeNode): Integer;
      function SlowFindRecord(const aNode: TTreeNode): Integer;
      procedure Sort;
    public
      //function Add(const aNode: TTreeNode; var aRec: tTransferToRec): tTransferToRec;
      function Delete(const aNode: TTreeNode): Boolean;
      procedure Add(const aNode: TTreeNode; var aRec: tTransferToRec);
      //procedure TransferRecord(const aNode: TTreeNode; var aRec: tTransferToRec);
      property TransferRec[const aNode: TTreeNode]: tTransferToRec read GetTransferRec;
  end;

  TReassignOrderRecordsHelper = record helper for tReassignOrderRecords
    procedure SortByNewProviderIEN;
  end;

  TOrderRecordsHelper = record helper for tOrderRecords
    procedure Delete(const Index: Cardinal);
    Procedure Add(const aSrc: tOrderRecord);
    Procedure SortByName;
    Procedure SortByOrderIEN;
    procedure SetSize(aValue: Integer);
  end;

  TStopWatch = class
  private
    FOwner: TObject;
    fFrequency: TLargeInteger;
    fIsRunning: Boolean;
    fIsHighResolution: Boolean;
    fStartCount, fStopCount: TLargeInteger;
    fActive: Boolean;
    procedure SetTickStamp(var lInt: TLargeInteger);
    function GetElapsedTicks: TLargeInteger;
    function GetElapsedMilliseconds: TLargeInteger;
    Function GetElapsedNanoSeconds: TLargeInteger;
    function GetElapsed: string;
  public
    constructor Create(AOwner: TComponent; const IsActive: Boolean = false;
      const startOnCreate: Boolean = false);
    destructor Destroy; override;
    procedure Start;
    procedure Stop;
    property IsHighResolution: Boolean read fIsHighResolution;
    property ElapsedTicks: TLargeInteger read GetElapsedTicks;
    property ElapsedMilliseconds: TLargeInteger read GetElapsedMilliseconds;
    property ElapsedNanoSeconds: TLargeInteger read GetElapsedNanoSeconds;
    property Elapsed: string read GetElapsed;
    property IsRunning: Boolean read fIsRunning;
  end;

  tProgressStatusBar = Class
    Private
      fShouldCancel: Boolean;
      fsbMain: TStatusBar;
      fPbMain: TProgressBar;
      fBtnCancel: TButton;
      fStopTimer: TStopWatch;
      fCrtLoop: Integer;
      fMaxLoops: Integer;
      fVisible: boolean;
      fCnRtnCsr: TCursor;
      fDisplayHint: boolean;
     // fProgDlg: TfrmProg;
      procedure DisplayHint(Sender: TObject);
      procedure UpdateEstimateTime;
      function GetVisible: Boolean;
      procedure SetVisible(const Value: Boolean);
      procedure OnCancelClick(Sender: TObject);
      procedure OnCancelMouseEnter(Sender: TObject);
      procedure OnCancelMouseLeave(Sender: TObject);
      function GetDisplayHints: Boolean;
      procedure SetDisplayHints(const Value: Boolean);
    Public
      constructor Create(aMainFrm: TForm);
      destructor Destroy(); override;
      procedure ResetProgressBar(aNewMax: Integer = 0);
      procedure ResetLoop(aNewMax: Integer);
      procedure HideProgress;
      procedure StepProgress;
      procedure AdjustPanelWidth(aPanelIdx: Integer; aAdjustBy: Integer);
      property Visible: Boolean read GetVisible write SetVisible;
      property CancelClicked: Boolean read fShouldCancel;
      property ReturnCursor: TCursor Read fCnRtnCsr write fCnRtnCsr;
      property DisplayHints: Boolean read GetDisplayHints write SetDisplayHints;
  End;

function FMNow: TFMDateTime;
function FMToday: TFMDateTime;
function OrderDetails(AOrderIEN: string; APatientIEN: string): string;
procedure GetTransferHistory(const AOrderIEN: string; var Dest: TStringList);
procedure SubSetOfProviders(const StartFrom: string; Direction: Integer;
  results: TStrings);
procedure SubSetOfActiveAndInactivePersons(const StartFrom: string; Direction: Integer;
  results: TStrings);
procedure SubSetOfPatients(const StartFrom: string; Direction: Integer;
  results: TStrings);
procedure AuditData(const Start, Stop: string; results: TStrings);
function ServerVersion(const Option, VerClient: string): string;
function ClientVersion(const AFileName: string): string;
function CompareVersion(const A, B: string): Integer;
procedure GetUserInfo;
function IsAuthUser: Boolean;

procedure ListSymbolTable(aDest: TStrings);
function GetUserParam(const AParamName: string): string;
procedure GetUserListParam(Dest: TStrings; const AParamName: string);
function StrToFMDateTime(const Value: String): TFMDateTime;
Function CompareInteger(const Value1, Value2: Integer): Integer;
Function CompareFloat(const Value1, Value2: Double): Integer;

Function PRTMessage(const CaptionText, TitleText, BodyText: String; const DlgButtons: array of String; const DefaultDlgButton: String; const DlgType: tTaskDLGType; const ExpandFoot: Boolean = false; const ExpandTitle: String = ''; const ExpandText: String = '' ): Integer;

//StatusBar & ProgressBar events

const
  MSG_CONNECT_TO_SERVER = WM_APP + 1;
  TX_OPTION = 'OR PRT GUI';
  // this is our "application context" ID for this application
  TC_CLIERR = 'Client Specifications Mismatch';

  grp_None = 0;
  grp_In = 1;
  grp_out = 2;
  grp_none_Header = 'Non-pending Order Transfers';
  grp_out_Header = 'Pending Order Transfers - Outgoing';
  grp_in_Header = 'Pending Order Transfers - Incoming';
  GrpCutoff = 0;  // Turn off grouping

var
  User: TUserInfo;
  XferSorted: Boolean;

implementation

uses
  vcl.Controls, System.Math, CommCtrl;

procedure GetUserInfo;
// stolen from CPRS uCore
{ returns a record of user information,
  Pieces: DUZ^NAME^USRCLS^CANSIGN^ISPROVIDER^ORDERROLE^NOORDER^DTIME^CNTDN^VERORD^NOTIFYAPPS^
  MSGHANG^DOMAIN^SERVICE^AUTOSAVE^INITTAB^LASTTAB^WEBACCESS^ALLOWHOLD^ISRPL^RPLLIST^
  CORTABS^RPTTAB^STATION#^GECStatus^Production account? }
var
  s: string;
begin
  CallVistA('ORWU USERINFO', [nil], s);
  with User do
  begin
    DUZ := StrToInt64Def(Piece(s, U, 1), 0);
    Name := Piece(s, U, 2);
    UserClass := StrToIntDef(Piece(s, U, 3), 0);
    CanSignOrders := Piece(s, U, 4) = '1';
    IsProvider := Piece(s, U, 5) = '1';
    OrderRole := StrToIntDef(Piece(s, U, 6), 0);
    NoOrdering := Piece(s, U, 7) = '1';
    DTIME := StrToIntDef(Piece(s, U, 8), 300);
    CountDown := StrToIntDef(Piece(s, U, 9), 10);
    EnableVerify := Piece(s, U, 10) = '1';
    EnableActOneStep := Piece(s, U, 27) = '0';
    NotifyAppsWM := Piece(s, U, 11) = '1';
    PtMsgHang := StrToIntDef(Piece(s, U, 12), 5);
    Domain := Piece(s, U, 13);
    Service := StrToIntDef(Piece(s, U, 14), 0);
    AutoSave := StrToIntDef(Piece(s, U, 15), 180);
    InitialTab := StrToIntDef(Piece(s, U, 16), 1);
    UseLastTab := Piece(s, U, 17) = '1';
    WebAccess := Piece(s, U, 18) <> '1';
    DisableHold := Piece(s, U, 19) = '1';
    IsRPL := Piece(s, U, 20);
    RPLList := Piece(s, U, 21);
    HasCorTabs := Piece(s, U, 22) = '1';
    HasRptTab := Piece(s, U, 23) = '1';
    StationNumber := Piece(s, U, 24);
    GECStatusCheck := Piece(s, U, 25) = '1';
    IsProductionAccount := Piece(s, U, 26) = '1';
    IsReportsOnly := false;
    if ((HasRptTab) and (not HasCorTabs)) then
      IsReportsOnly := true;
    // Remove next if and nested if should an "override" later be provided for RPL users,etc.:
    if HasCorTabs then
      if (IsRPL = '1') then
      begin
        IsRPL := '0'; // Hard set for now.
        IsReportsOnly := false;
      end;
    // Following hard set to TRUE per VHA mgt decision:
    ToolsRptEdit := true;
    // x := GetUserParam('ORWT TOOLS RPT SETTINGS OFF');
    // if x = '1' then
    // ToolsRptEdit := false;
  end;
end;

function FMNow: TFMDateTime;
// Borrowed from RCore. Return the current date/time in Fileman format }
begin
  Result := GetFMNow;
end;

function FMToday: TFMDateTime;
{ return the current date in Fileman format }
begin
  Result := Int(FMNow);
end;

function OrderDetails(AOrderIEN: string; APatientIEN: string): string;
// returns order details
var
  results: TStringList;
begin
  results := TStringList.Create;
  try
    CallVistA('ORQOR DETAIL', [AOrderIEN, APatientIEN], results);
    Result := results.Text;
  finally
    freeAndNil(results);
  end;
end;

procedure GetTransferHistory(const AOrderIEN: string; var Dest: TStringList);
begin
  if Assigned(Dest) then
  begin
    Dest.Clear;
    CallVistA('ORQ3 XFER HISTORY', [AOrderIEN], Dest);
  end;
end;

procedure SubSetOfProviders(const StartFrom: string; Direction: Integer;
  results: TStrings);
// SDS: Stolen and modified from CPRS rCore.pas
// returns a pointer to a list of providers (for use in a long list box) -
begin
  CallVistA('ORWU NEWPERS', [StartFrom, Direction, 'PROVIDER'], results);
end;

procedure SubSetOfActiveAndInactivePersons(const StartFrom: string; Direction: Integer;
  results: TStrings);
// SDS: Stolen and modified from CPRS rCore.pas
// returns a pointer to a list of providers (for use in a long list box) -
begin
  CallVistA('ORWU NEWPERS', [StartFrom, Direction, 'PROVIDER', '', '', true], results);
end;

procedure AuditData(const Start, Stop: string; results: TStrings);
// returns a list of providers or patients with transfer history
begin
  CallVista('ORQ3 AUDIT', [Start, Stop], results);
end;

procedure SubSetOfPatients(const StartFrom: string; Direction: Integer;
  results: TStrings);
{ returns a pointer to a list of patients (for use in a long list box) -  The return value is
  a pointer to RPCBrokerV.Results, so the data must be used BEFORE the next broker call! }
begin
  CallVistA('ORWPT LIST ALL', [StartFrom, Direction], results);
  MixedCaseList(results);
end;

function ServerVersion(const Option, VerClient: string): string;
// get the server version from Vista
begin
  CallVistA('ORWU VERSRV', [Option, VerClient], Result);
end;

function ClientVersion(const AFileName: string): string;
// get the GUI version information stored in Project Options VerInfo
var
  ASize, AHandle: DWORD;
  Buf: string;
  FileInfoPtr: Pointer; // PVSFixedFileInfo;
begin
  Result := '';
  ASize := GetFileVersionInfoSize(PChar(AFileName), AHandle);
  if ASize > 0 then
  begin
    SetLength(Buf, ASize);
    GetFileVersionInfo(PChar(AFileName), AHandle, ASize, Pointer(Buf));
    VerQueryValue(Pointer(Buf), '\', FileInfoPtr, ASize);
    with TVSFixedFileInfo(FileInfoPtr^) do
      Result := IntToStr(HIWORD(dwFileVersionMS)) + '.' +
        IntToStr(LOWORD(dwFileVersionMS)) + '.' +
        IntToStr(HIWORD(dwFileVersionLS)) + '.' +
        IntToStr(LOWORD(dwFileVersionLS));
  end;
end;

function CompareVersion(const A, B: string): Integer;
var
  NumA, NumB: Integer;
begin
  NumA := (StrToInt(Piece(A, '.', 1)) * 16777216) +
    (StrToInt(Piece(A, '.', 2)) * 65536) + (StrToInt(Piece(A, '.', 3)) * 256) +
    StrToInt(Piece(A, '.', 4));
  NumB := (StrToInt(Piece(B, '.', 1)) * 16777216) +
    (StrToInt(Piece(B, '.', 2)) * 65536) + (StrToInt(Piece(B, '.', 3)) * 256) +
    StrToInt(Piece(B, '.', 4));
  Result := NumA - NumB;
end;

function IsAuthUser: Boolean;
// We call ORWU TOOLMENU and verify it contains the text "ProviderRoleTool.exe"
// We do this check in case the user is trying to run us from the command line
// note that if you're running the debug version, I just let you in
{$IFNDEF DEBUG}
var
  s: string;
{$ENDIF}
begin
{$IFDEF DEBUG}
  Result := true;
{$ELSE}
  CallVistA('ORQ3 AUTHUSR', [], s);
  Result := s = '1';
{$ENDIF}
end;

procedure ListSymbolTable(aDest: TStrings);
var
  results: TStrings;
  i: Integer;
  x: string;
begin
  aDest.Clear;
  results := TStringList.Create;
  try
    CallVistA('ORWUX SYMTAB', [nil], results);
    i := 0;
    with results do
      while i < Count do
      begin
        x := Strings[i] + '=';
        Inc(i);
        if i < Count then
          x := x + Strings[i];
        aDest.Add(x);
        Inc(i);
      end;
  finally
    results.Free;
  end;
end;

function GetUserParam(const AParamName: string): string;
begin
  CallVistA('ORWU PARAM', [AParamName], Result);
end;

procedure GetUserListParam(Dest: TStrings; const AParamName: string);
begin
  CallVistA('ORWU PARAMS', [AParamName], Dest);
end;

function StrToFMDateTime(const Value: String): TFMDateTime;
begin
  Result := 0.0;
  if Value <> '' then
    Result := DateTimeToFMDateTime(StrToDateTime(Value))
end;

Function CompareInteger(const Value1, Value2: Integer): Integer;
begin
 Result := CompareFloat(Value1, Value2);
end;

Function CompareFloat(const Value1, Value2: Double): Integer;
begin
 if Value1 < Value2 then
    Result := -1
  else if Value2 < Value1 then
    Result := 1
  else
    Result := 0;
end;

Function PRTMessage(const CaptionText, TitleText, BodyText: String;
  const DlgButtons: array of String; const DefaultDlgButton: String;
  const DlgType: tTaskDLGType; const ExpandFoot: Boolean = false;
  const ExpandTitle: String = ''; const ExpandText: String = ''): Integer;
var
  i, x: Integer;
  aMsgdlg: TForm;
  aMsgDlgType: TMsgDlgType;
  aMsgButtons: TMsgDlgButtons;
  aTskDlg: TTaskDialog;
  aTskDlgBtn: TTaskDialogButtonItem;
  Btn: TButton;
  BtnIdx: Integer;
begin
  Result := -1;

  if (Length(DlgButtons) = 0) or (Length(DlgButtons) > 12) then
    exit;

  if (Win32MajorVersion >= 6) then
  begin
{$WARN SYMBOL_PLATFORM OFF}
    aTskDlg := TTaskDialog.Create(nil);
    try
      aTskDlg.Caption := CaptionText;
      aTskDlg.Title := TitleText;
      aTskDlg.Text := BodyText;
      aTskDlg.CommonButtons := [];
      for i := Low(DlgButtons) to high(DlgButtons) do
      begin
        aTskDlgBtn := TTaskDialogButtonItem(aTskDlg.Buttons.Add);
        aTskDlgBtn.Caption := DlgButtons[i];
        aTskDlgBtn.default := DlgButtons[i] = DefaultDlgButton;
        aTskDlgBtn.ModalResult := i;
      end;

      if ExpandFoot then
        aTskDlg.Flags := [tfAllowDialogCancellation, tfExpandFooterArea]
      else
        aTskDlg.Flags := [tfAllowDialogCancellation];

      aTskDlg.ExpandButtonCaption := ExpandTitle;
      aTskDlg.ExpandedText := ExpandText;
      case DlgType of
        tdlgNone:
          aTskDlg.MainIcon := tdiNone;
        tdlgWarning:
          aTskDlg.MainIcon := tdiWarning;
        tdlgError:
          aTskDlg.MainIcon := tdiError;
        tdlgInformation:
          aTskDlg.MainIcon := tdiInformation;
        tdlgShield:
          aTskDlg.MainIcon := tdiShield;
      end;

      if aTskDlg.Execute then
        Result := aTskDlg.ModalResult;

    finally
      aTskDlg.Free;
    end
{$WARN SYMBOL_PLATFORM ON}
  end
  else
  begin

    case DlgType of
      tdlgNone:
        aMsgDlgType := mtCustom;
      tdlgWarning:
        aMsgDlgType := mtWarning;
      tdlgError:
        aMsgDlgType := mtError;
      tdlgInformation:
        aMsgDlgType := mtInformation;
      tdlgShield:
        aMsgDlgType := mtConfirmation;
    else
      aMsgDlgType := mtConfirmation;
    end;

    aMsgButtons := [];
    for i := Low(DlgButtons) to high(DlgButtons) do
    begin
      Include(aMsgButtons, TMsgDlgBtn(i));
    end;

    aMsgdlg := createMessageDialog(BodyText, aMsgDlgType, aMsgButtons);
    aMsgdlg.Caption := BodyText;
    aMsgdlg.BiDiMode := bdRightToLeft;

    BtnIdx := 0;
    for i := Low(DlgButtons) to High(DlgButtons) do
    begin
      for x := BtnIdx to aMsgdlg.componentcount - 1 Do
      begin
        if (aMsgdlg.components[x] is vcl.StdCtrls.TButton) then
        Begin
          Btn := TButton(aMsgdlg.components[x]);
          Btn.Caption := DlgButtons[i];
          BtnIdx := x + 1;
          break;
        End;
      end;
    end;

    Result := aMsgdlg.Showmodal;
  end;

end;

{$REGION 'Helpers'}

{$REGION 'THelperListGroup'}
procedure THelperListGroup.IncFooterCount(PieceNum: Integer);
var
 TempNum: Integer;
 tempStr: String;
begin
 TempNum := StrToIntDef(Piece(self.Footer, ' ' , PieceNum),0);
 Inc(TempNum);
 tempStr := self.Footer;
 SetPiece(tempStr, ' ' , PieceNum, IntToStr(TempNum));
 self.Footer := tempStr;
end;

function THelperListGroups.FindItemByGroupID(GroupID: Integer): TListGroup;
var
  i: Integer;
begin
  for i := 0 to Count - 1 do
  begin
    Result := Items[i];
    if Result.GroupID = GroupID then
      Exit;
  end;
  Result := nil;
end;

function THelperListGroups.FindItemByGroupHeader(GroupHeader: String)
  : TListGroup;
var
  i: Integer;
begin
  for i := 0 to Count - 1 do
  begin
    Result := Items[i];
    if trim(Result.Header) = trim(GroupHeader) then
      Exit;
  end;
  Result := nil;
end;

function THelperListGroups.FindItemByGroup(aLookup: String;
  const SrchDict: tSrchDictGroup): TListGroup;
begin
  Result := nil;
  SrchDict.TryGetValue(aLookup, Result);
end;

function THelperListGroups.FindLastGroupID: Integer;
var
  i: Integer;
begin
  Result := 0;
  for i := Count - 1 downto 0 do
  begin
    if Items[i].GroupID > Result then
      Result := Items[i].GroupID;
  end;

end;
{$ENDREGION}

{$REGION 'tStringlistHelper'}

function tStringlistHelper.IndexOfPiece(const S: string; const PieceDelim: Char; const PieceNum: integer): Integer;
begin
 Result := IndexOfPieceEx(S, PieceDelim, PieceNum, 0);
end;

function tStringlistHelper.IndexOfPieceEx(const S: string; const PieceDelim: Char; const PieceNum: integer; const StartFrom: Integer): Integer;
var
  Count: Integer;
  SLen: Integer;
  aStr: String;
begin
  Count := GetCount;

  if StartFrom > count then
  begin
    Result := -1;
    exit;
  end;

  if not Sorted then
  begin

    SLen := Length(S);
    for Result := StartFrom to Count - 1 do
    begin
      aStr := Piece(self.Strings[Result], PieceDelim, PieceNum);
      if (Length(aStr) = SLen) and (CompareStrings(aStr, S) = 0) then
        Exit;
    end;
    Result := -1;
  end
  else
    if not FindByPieceEx(S, Result, PieceDelim, PieceNum, StartFrom) then
      Result := -1;
end;

function tStringlistHelper.FindByPiece(const S: string; var Index: Integer; const PieceDelim: Char; const PieceNum: integer): Boolean;
begin
  Result := FindByPieceEx(S, Index, PieceDelim, PieceNum, 0);
end;

function tStringlistHelper.FindByPieceEx(const S: string; var Index: Integer; const PieceDelim: Char; const PieceNum: integer; const StartFrom: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;
  L := 0;
  H := GetCount - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := CompareStrings(Piece(self.Strings[I], PieceDelim, PieceNum), S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        if Duplicates <> dupAccept then L := I;
      end;
    end;
  end;
  Index := L;
end;

{$ENDREGION}

{$REGION 'THelperListView'}

function THelperListView.FindListItemByPatient(const aIEN: string;
  const SrchDict: tSrchDict): TListItem;
begin
  result := nil;
  SrchDict.TryGetValue(aIEN, result);
end;

function THelperListView.FindListItemByXfer(const aIEN: string;
  const XferType: tXfertType; const SrchDict: tSrchDict): TListItem;
begin
  result := nil;

  case XferType of
    txftPendOut:
      SrchDict.TryGetValue(IntToStr(grp_out) + '_' + aIEN, result);
    txftPendIn:
      SrchDict.TryGetValue(IntToStr(grp_In) + '_' + aIEN, result);
    else
      SrchDict.TryGetValue(IntToStr(grp_None) + '_' + aIEN, result);
  end;

end;

Function THelperListView.AddListItemGroupByName(const aOrder: tOrderRecord;
  const SrchDict: tSrchDictGroup; CreatsGroups: Boolean): TListItem;
var
  Group: TListGroup;
  aListObj: pListObj;
begin

  System.New(aListObj);
  // Each individual order is added here
  aListObj.Group_Type := tvtOrder;
  aListObj.OrderIEN := aOrder.OrderIEN;
  aListObj.NumOfOrders := 1;
  aListObj.patientIEN := aOrder.patient.IEN;

  result := Items.Add;
  if CreatsGroups then
  begin
    Group := Groups.FindItemByGroup(self.name + '_' + aListObj.patientIEN,
      SrchDict);

    if not assigned(Group) then
    begin
      Group := Groups.Add;
      Group.GroupID := StrToIntDef(aOrder.patient.IEN, 0);
      Group.Header := aOrder.patient.name;
      Group.State := [lgsCollapsible];
   //   Group.Footer := '0 orders';
      SrchDict.Add(self.name + '_' + aOrder.patient.IEN, Group);
    end;
  //  Group.IncFooterCount(1);
    result.GroupID := Group.GroupID;
  end;

  result.Caption := aOrder.OrderIEN;
  result.SubItems.Add(aOrder.orderDate);
  result.SubItems.Add(aOrder.orderStatus);
  result.SubItems.Add(FormatFMDateTimeStr('mm/dd/yy@hh:nn', aOrder.orderXferEffective));
  result.SubItems.Add(aOrder.orderXferFrom.name);
  if not CreatsGroups then
    result.SubItems.Add(aOrder.patient.name);
  result.Data := aListObj;

end;

Function THelperListView.AddListItemGroupByXfer(const aOrder: tOrderRecord;
  const SrchDict: tSrchDict): TListItem;
var
  Group: TListGroup;
  aListObj: pListObj;
  tmpGrpID: Integer;
  TmpGrpTxt: String;
begin
  System.New(aListObj);
  // since this is all the orders for this patient and xfer type then we only
  // need the first order for the lookup
  aListObj.Group_Type := tvtXfer;
  aListObj.OrderIEN := aOrder.OrderIEN;
  aListObj.NumOfOrders := 1;
  aListObj.patientIEN := aOrder.patient.IEN;

  GroupView := true;

  case aOrder.OrderXfer of
    txftPendOut:
      begin
        tmpGrpID := grp_out;
        TmpGrpTxt := grp_out_Header;
      end;
    txftPendIn:
      begin
        tmpGrpID := grp_In;
        TmpGrpTxt := grp_in_Header;
      end;
    else
    begin
      tmpGrpID := grp_None;
      TmpGrpTxt := grp_none_Header;
    end;
  end;

  if Groups.Count < 3 then
  begin
    Group := Groups.FindItemByGroupID(tmpGrpID);
    if not assigned(Group) then
    begin
      Group := Groups.Add;
      Group.State := [lgsCollapsible];
      Group.GroupID := tmpGrpID;
      Group.Header := TmpGrpTxt;
    end;
  end;

  result := Items.Add;
  result.GroupID := tmpGrpID;
  result.Caption := aOrder.patient.name;
  result.SubItems.Add('1');
  result.Data := aListObj;

  SrchDict.Add(IntToStr(tmpGrpID) + '_' + aListObj.patientIEN, result);
end;

Function THelperListView.AddListItemGroupByPatient(const aOrder: tOrderRecord;
  const SrchDict: tSrchDict): TListItem;
var
  aListObj: pListObj;
begin
  System.New(aListObj);
  // since this is all the orders for this patient then we only need the first
  // order for the lookup
  aListObj.Group_Type := tvtPatient;
  aListObj.OrderIEN := aOrder.OrderIEN;
  aListObj.NumOfOrders := 1;
  aListObj.patientIEN := aOrder.patient.IEN;

  result := Items.Add;
  result.Caption := aOrder.patient.name;
  result.SubItems.Add('1');
  result.Data := aListObj;

  SrchDict.Add(aListObj.patientIEN, result);
end;

Function THelperListView.CleanUpListItem(const aItem: TListItem): Boolean;
begin
  result := false;
  if assigned(aItem.Data) then
  begin
    System.Dispose(pListObj(aItem.Data));
    result := true;
  end;
end;

Procedure THelperListView.IncOrderCount(const aItem: TListItem;
  IncBy: Integer = 1);
Var
  aListObj: pListObj;
  TmpCnt: Integer;
begin
  if assigned(aItem.Data) then
  begin
    aListObj := pListObj(aItem.Data);
    TmpCnt := StrToIntDef(aItem.SubItems[0], 0);
    inc(TmpCnt, IncBy);
    aListObj.NumOfOrders := TmpCnt;
    aItem.SubItems[0] := IntToStr(TmpCnt);
  end;
end;

{$ENDREGION}

{$REGION 'TReassignOrderRecordsHelper'}

 procedure TReassignOrderRecordsHelper.SortByNewProviderIEN;
begin
  // sort the array by New provider IEN
  TArray.Sort<tReassignOrderRecord>(self,
    TDelegatedComparer<tReassignOrderRecord>.Construct(
    function(const Left, Right: tReassignOrderRecord): Integer
    begin
      Result := TComparer<String>.Default.Compare(Left.newProviderIEN,
        Right.newProviderIEN);
    end));
end;

{$ENDREGION}

{$REGION 'TtOrderRecordsHelper'}

procedure TOrderRecordsHelper.Delete(const Index: Cardinal);
var
  ALength: Cardinal;
  TailElements: Cardinal;
begin
  ALength := Length(Self);
  Assert(ALength > 0);
  Assert(Index < ALength);
  Finalize(Self[Index]);
  TailElements := ALength - Index;
  if TailElements > 0 then
    Move(Self[Index + 1], Self[Index], SizeOf(tOrderRecord) * TailElements);
  Initialize(self[ALength - 1]);
  SetLength(self, ALength - 1);
end;

Procedure TOrderRecordsHelper.Add(const aSrc: tOrderRecord);
begin
  SetLength(self, Length(self) + 1);
  self[high(self)] := aSrc;
//  self[high(self)].Assign(aSrc);
end;

procedure TOrderRecordsHelper.SetSize(aValue: Integer);
begin
  SetLength(self, aValue);
end;


Procedure TOrderRecordsHelper.SortByName;
begin
    TArray.Sort<tOrderRecord>(Self,
      TDelegatedComparer<tOrderRecord>.Construct(
      function(const Left, Right: tOrderRecord): integer
      begin
          result := TComparer<String>.Default.Compare(Left.patient.name,
            Right.patient.name)
      end));
end;

Procedure TOrderRecordsHelper.SortByOrderIEN;
begin
    TArray.Sort<tOrderRecord>(Self,
      TDelegatedComparer<tOrderRecord>.Construct(
      function(const Left, Right: tOrderRecord): integer
      begin
          result := TComparer<Integer>.Default.Compare(StrToIntDef(Left.orderIEN, -1),
            StrToIntDef(Right.orderIEN, -1))
      end));
end;

{$ENDREGION}

{$REGION 'TTransferOdersHelper'}

{
procedure TTransferOdersHelper.DeleteIdx(const Index: Cardinal);
var
  ALength: Cardinal;
  TailElements: Cardinal;
begin
  ALength := Length(Self);
  Assert(ALength > 0);
  Assert(Index < ALength);
  Finalize(Self[Index]);
  TailElements := ALength - Index;
  if TailElements > 0 then
    Move(Self[Index + 1], Self[Index], SizeOf(tTransferToRec) * TailElements);
  Initialize(self[ALength - 1]);
  SetLength(self, ALength - 1);
end;
}
procedure TTransferOdersHelper.Sort;
begin
  if not XferSorted  then
  begin
    TArray.Sort<tTransferToRec>(self,
      TDelegatedComparer<tTransferToRec>.Construct(
      function(const Left, Right: tTransferToRec): integer
      begin
        result := TComparer<Integer>.Default.Compare(left.NodePointer, Right.NodePointer);
      end));

    XferSorted := true;
  end;
end;

function TTransferOdersHelper.FindRecord(const aNode: TTreeNode): Integer;
var
 FndIdx: Integer;
 aXfer: tTransferToRec;
 Found: Boolean;
begin
  Result := -1;
  //Sort by IEN to speed up the search
  Sort;

  //Perform the search
  aXfer.NodePointer := NativeInt(aNode);
  Found:= TArray.BinarySearch<tTransferToRec>(self, aXfer, FndIdx,
    TDelegatedComparer<tTransferToRec>.Construct(
    function(const Left, Right: tTransferToRec): Integer
    begin
      if Left.NodePointer > Right.NodePointer then Exit(1);
      if Left.NodePointer < Right.NodePointer then Exit(-1);
      Result:= 0;  //or raise an error if duplicates are not allowed.
    end
  ));
  if Found then
  begin
     //Fail safe
     if self[FndIdx].aTreeNode = aNode then
      result := FndIdx
     else
      SlowFindRecord(aNode);
  end else
    SlowFindRecord(aNode);
end;

function TTransferOdersHelper.SlowFindRecord(const aNode: TTreeNode): Integer;
var
 I: Integer;
begin
 Result := -1;
  for I := Low(self) to High(self) do
  begin
    if self[i].aTreeNode = aNode then
    begin
     Result := I;
     break;
    end;
  end;
end;

procedure TTransferOdersHelper.Add(const aNode: TTreeNode;
  var aRec: tTransferToRec);
begin
 SetLength(Self, Length(self) + 1);
 Self[high(self)] := aRec;
 Self[high(self)].aTreeNode := aNode;
 Self[high(self)].NodePointer := NativeInt(aNode);//Integer(Pointer(aNode));
 XferSorted := false;
end;

function TTransferOdersHelper.Delete(const aNode: TTreeNode): Boolean;
var
 DelIdx: Integer;
begin
  Result := false;
  DelIdx := FindRecord(aNode);
  if DelIdx <> -1 then
  begin
    system.Delete(self, DelIdx, 1);
    result := true;
  end;
end;

function TTransferOdersHelper.GetTransferRec(
  const aNode: TTreeNode): tTransferToRec;
Var
  FndIdx: Integer;
begin
 FndIdx := FindRecord(aNode);
 if FndIdx > -1 then
  Result := Self[FndIdx];
end;

{$ENDREGION}

{$REGION 'ttFilterRecHelper'}

function ttFilterRecHelper.GetFilterRec(const aIdx: Integer): tFilterRec;
var
  ArIdx: Integer;
begin
  ArIdx := GetFilterRecIndex(aIdx);

  if ArIdx <> -1 then
    result := self[ArIdx];

end;

function ttFilterRecHelper.GetFilterRecIndex(const aIdx: Integer; const AutoAdd: Boolean = false; const aValue: String = ''; const aChecked: Boolean = false): Integer;
var
  I: integer;
begin
  Result := -1;
  for I := Low(self) to High(self) do
  begin
    if self[i].Index = aIdx then
    begin
     Result := I;
     break;
    end;
  end;

  if AutoAdd and (Result = -1) then
  begin
    SetLength(self, length(self) + 1);
    Result := High(self);
    self[Result].Index := aIdx;
    self[Result].Value := aValue;
    self[Result].InitalChecked := aChecked;
  end;
end;

function ttFilterRecHelper.FilterRecordIndex(const aIdx: Integer; const AutoAdd: Boolean = false; const aValue: String = ''; const aChecked: Boolean = false): Integer;
begin
 Result := GetFilterRecIndex(aIdx, AutoAdd, aValue, aChecked);
end;

procedure ttFilterRecHelper.Clear;
begin
  SetLength(Self, 0);
end;
{
procedure ttFilterRecHelper.DeleteIdx(const Index: Cardinal);
var
  ALength: Cardinal;
  TailElements: Cardinal;
begin
  ALength := Length(Self);
  Assert(ALength > 0);
  Assert(Index < ALength);
  Finalize(Self[Index]);
  TailElements := ALength - Index;
  if TailElements > 0 then
    Move(Self[Index + 1], Self[Index], SizeOf(tTransferToRec) * TailElements);
  Initialize(self[ALength - 1]);
  SetLength(self, ALength - 1);
end;
}
function ttFilterRecHelper.Delete(const aIdx: Integer): Boolean;
begin
  Result := true;
  system.Delete(self, aIdx, 1);
end;


{$ENDREGION}

{$ENDREGION}

{$REGION 'Records'}

{$REGION 'tPersonRecord'}
procedure tPersonRecord.Clear;
// clears the record
begin
  tag := 0;
  name := '';
  IEN := '0';
  IsProvider := false;
end;

procedure tPersonRecord.Initialize(aName: string; aIEN: string;
aIsProivder: Boolean);
// initializes record to passed values
begin
  name := aName;
  IEN := aIEN;
  IsProvider := aIsProivder;
end;

function tPersonRecord.ToNodeName: string;
begin
  if not IsProvider then
    result := format('%s (%d)', [Name, tag])
  else
    result := format('%s', [Name])
end;

function tPersonRecord.ToStringData: string;
begin
  result := IEN + U + Name + U + BoolToStr(IsProvider);
end;

Procedure tPersonRecord.Assign(const aSrc: tPersonRecord);
begin
  IEN := aSrc.IEN;
  name := aSrc.name;
  tag := aSrc.tag;
  IsProvider := aSrc.IsProvider;
  OrderXfer := aSrc.OrderXfer;
end;

function tPersonRecord.Valid: Boolean;
// is the data in this record valid?
begin
  result := IEN <> '0';
end;
{$ENDREGION}

{$REGION 'tOrderRecord'}
Procedure tOrderRecord.Assign(const aSrc: tOrderRecord);
begin
  patient.Assign(aSrc.patient);
  orderIEN := aSrc.orderIEN;
  orderStatus := aSrc.orderStatus;
  orderDate := aSrc.orderDate;
  OrderXfer := aSrc.OrderXfer;
  orderXferEffective := aSrc.orderXferEffective;
  orderXferFrom := aSrc.orderXferFrom;
  orderFlag := aSrc.orderFlag;
end;

function tOrderRecord.CompareRecord(const aCompare: tOrderRecord): Boolean;
begin
  // Can not have the same order IEN
  result := (Self.orderIEN = aCompare.orderIEN)
end;

Procedure tOrderRecord.LoadFromString(const aValue: String);
var
  tmpxfer: integer;
begin
  patient.name := piece(aValue, '^', 1);
  patient.IEN := piece(aValue, '^', 2);
  orderIEN := piece(aValue, '^', 3);
  orderStatus := piece(aValue, '^', 4);
  orderDate := piece(aValue, '^', 5);
  // blank (0) = none, 1 = Transfer out (pending transfer), 2 = Order has been Transferred to this provider, 3 = Transferred
  tmpxfer := StrToIntDef(piece(aValue, '^', 6), 0);
  case tmpxfer of
    0:
      OrderXfer := txftNone;
    1:
      OrderXfer := txftPendOut;
    2:
      OrderXfer := txftPendIn;
    3:
      OrderXfer := txftTransferred;
  end;
  orderXferEffective := piece(aValue, '^', 7);
  orderXferFrom.name := piece(aValue, '^', 8);
  orderXferFrom.IEN := piece(aValue, '^', 9);
  orderFlag := tofNA;
end;
{$ENDREGION}

{$REGION 'tTransferToRec'}

Procedure tTransferToRec.Assign(const aSrc: tTransferToRec);
begin
  PatientInfo.Assign(aSrc.PatientInfo);
 // ListObject.Assign(aSrc.ListObject);
  IsProviderDisplay := aSrc.IsProviderDisplay;
end;

Procedure tListObject.Assign(const aSrc: tListObject);
begin
  Group_Type := aSrc.Group_Type;
  orderIEN := aSrc.orderIEN;
  NumOfOrders := aSrc.NumOfOrders;
end;


{$ENDREGION}

{$ENDREGION}

{$REGION 'Objects'}

{$REGION 'TStopWatch'}

Const
  NSecsPerSec = 1000000000;

constructor TStopWatch.Create(AOwner: TComponent;
const IsActive: Boolean = false; const startOnCreate: Boolean = false);
begin
  inherited Create();
  FOwner := AOwner;
  fIsRunning := false;
  fActive := IsActive;
  fIsHighResolution := QueryPerformanceFrequency(fFrequency);
  if NOT fIsHighResolution then
    fFrequency := MSecsPerSec;

  if startOnCreate then
    Start;
end;

destructor TStopWatch.Destroy;
begin
  Stop;
  inherited Destroy;
end;

function TStopWatch.GetElapsedTicks: TLargeInteger;
begin
  Result := fStopCount - fStartCount;
end;

procedure TStopWatch.SetTickStamp(var lInt: TLargeInteger);
begin
  if fIsHighResolution then
    QueryPerformanceCounter(lInt)
  else
    lInt := MilliSecondOf(Now);
end;

function TStopWatch.GetElapsed: string;
begin
  Result := FloatToStr(ElapsedMilliseconds / 1000) + ' Sec / ' +
    FloatToStr(ElapsedMilliseconds) + ' Ms / ' +
    FloatToStr(ElapsedNanoSeconds) + ' Ns';
end;

function TStopWatch.GetElapsedMilliseconds: TLargeInteger;
var
  Crnt: TLargeInteger;
begin
  if fIsRunning then
  begin
    SetTickStamp(Crnt);
    Result := (MSecsPerSec * (Crnt - fStartCount)) div fFrequency;
  end
  else
    Result := (MSecsPerSec * (fStopCount - fStartCount)) div fFrequency;
end;

function TStopWatch.GetElapsedNanoSeconds: TLargeInteger;
begin
  Result := (NSecsPerSec * (fStopCount - fStartCount)) div fFrequency;
end;

procedure TStopWatch.Start;
begin
  if fActive then
  begin
    SetTickStamp(fStartCount);
    fIsRunning := true;
  end;
end;

procedure TStopWatch.Stop;
begin
  if fActive then
  begin
    SetTickStamp(fStopCount);
    fIsRunning := false;
  end;
end;
{$ENDREGION}

{$REGION 'tProgressStatusBar'}
constructor tProgressStatusBar.Create(aMainFrm: TForm);
var
  NewStyle: Integer;
begin
  inherited Create();

  fsbMain := TStatusBar.Create(aMainFrm);
  fPbMain := TProgressBar.Create(aMainFrm);
  fBtnCancel := TButton.Create(aMainFrm);

  fsbMain.Parent := aMainFrm;
  fsbMain.Align := alBottom;
  fsbMain.Top := aMainFrm.Height - 1;

  with fsbMain.Panels.Add do
    Width := 200;
  with fsbMain.Panels.Add do
    Width := 50;
  with fsbMain.Panels.Add do
    Width := 200;

  fBtnCancel.Caption := 'Cancel';
  fBtnCancel.OnClick := OnCancelClick;
  fCnRtnCsr := crDefault;
  fBtnCancel.OnMouseEnter := OnCancelMouseEnter;
  fBtnCancel.OnMouseLeave := OnCancelMouseLeave;

  SetWindowLong(fsbMain.Handle, GWL_EXSTYLE, GetWindowLong(fsbMain.Handle,
    GWL_EXSTYLE) or WS_EX_COMPOSITED);

  fPbMain.parent := fsbMain;
  NewStyle := GetWindowLong(fPbMain.Handle, GWL_EXSTYLE);
  NewStyle := NewStyle and not WS_EX_STATICEDGE;
  SetWindowLong(fPbMain.Handle, GWL_EXSTYLE, NewStyle);

  fbtnCancel.parent := fsbMain;
  NewStyle := GetWindowLong(fbtnCancel.Handle, GWL_EXSTYLE);
  NewStyle := NewStyle and WS_EX_COMPOSITED and not WS_EX_STATICEDGE;
  SetWindowLong(fbtnCancel.Handle, GWL_EXSTYLE, NewStyle);
  fShouldCancel := false;

  fStopTimer := TStopWatch.Create(Application, true);
  //fProgDlg := TfrmProg.Create(nil);
  end;

destructor tProgressStatusBar.Destroy;
begin
  FreeAndNil(fStopTimer);
  FreeAndNil(fBtnCancel);
  FreeAndNil(fPbMain);
  FreeAndNil(fsbMain);
 // freeandnil(fProgDlg);
  Inherited;
end;

Procedure tProgressStatusBar.ResetProgressBar(aNewMax: Integer = 0);
begin
  fPbMain.Min := 0;
  fPbMain.Max := aNewMax;
  fPbMain.Position := 0;
  fPbMain.step := 1;
  fPbMain.visible := true;

  {fProgDlg.ProgressBar1.Min := 0;
  fProgDlg.ProgressBar1.Max := aNewMax;
  fProgDlg.ProgressBar1.Position := 0;
  fProgDlg.ProgressBar1.step := 1;
  fProgDlg.Show;    }


  fShouldCancel := false;
  fBtnCancel.visible := true;
  //fBtnCancel.SetFocus;
end;

procedure tProgressStatusBar.ResetLoop(aNewMax: Integer);
begin
 fCrtLoop := 0;
 fMaxLoops := aNewMax;
 //fProgDlg.CurrentLoop := 0;
// fProgDlg.MaxLoop := aNewMax;
end;

procedure tProgressStatusBar.HideProgress;
begin
  fBtnCancel.visible := false;
  fPbMain.visible := false;
 // fProgDlg.Hide;

end;

procedure tProgressStatusBar.DisplayHint(Sender: TObject);
begin
  fsbMain.Panels[0].Text := GetLongHint(Application.Hint);
end;

function tProgressStatusBar.GetDisplayHints: Boolean;
begin
 Result := fDisplayHint;
end;

function tProgressStatusBar.GetVisible: Boolean;
begin
 result := fVisible;
end;

Procedure tProgressStatusBar.UpdateEstimateTime;
var
  atmp: Double;
begin
  if not fStopTimer.IsRunning then
    fStopTimer.Start;
  if (fCrtLoop mod 10) = 1 then
  begin
    atmp := ((fStopTimer.ElapsedMilliseconds / 1000) / fCrtLoop) *
      (fMaxLoops - fCrtLoop);

    fsbMain.Panels[0].Text := 'Estimated ' + FormatDateTime('hh:nn:ss',
      System.Math.ceil(atmp) / SecsPerDay) + ' left ';
    { - ' + IntToStr(CurrentLoop)
      + ' of ' + IntToStr(MaxLoops); }

    // sleep(0);
    Application.ProcessMessages;
  end;
  if fCrtLoop >= fMaxLoops then
  begin
    fsbMain.Panels[0].Text := '';
    fStopTimer.Stop;
  end;

end;

procedure tProgressStatusBar.SetDisplayHints(const Value: Boolean);
begin
 fDisplayHint := Value;
 Application.OnHint := DisplayHint;
end;

procedure tProgressStatusBar.SetVisible(const Value: Boolean);
begin
 fVisible := Value;
 fsbMain.Visible := fVisible;
 if fVisible then
 begin
  fsbMain.Align := alBottom;
  fsbMain.Top := fsbMain.Parent.Height - 1;
  HideProgress;
 end;
end;

Procedure tProgressStatusBar.StepProgress;
begin
  Inc(fCrtLoop);
  fPbMain.StepIt;

  //fProgDlg.CurrentLoop := fCrtLoop;
 // fProgDlg.ProgressBar1.StepIt;

  application.ProcessMessages;
  UpdateEstimateTime;
end;

procedure tProgressStatusBar.AdjustPanelWidth(aPanelIdx: Integer; aAdjustBy: Integer);
var
  r: TRect;
begin
   fsbMain.Panels[aPanelIdx].Width := fsbMain.Width +
    aAdjustBy;

  // place the progress bar over the 2nd panel
  SendMessage(fsbMain.Handle, SB_GETRECT, 2, LPARAM(@r));
  fPbMain.SetBounds(r.Left, r.Top, r.Width, r.Height);

  SendMessage(fsbMain.Handle, SB_GETRECT, 1, LPARAM(@r));
  fbtnCancel.SetBounds(r.Left, r.Top, r.Width, r.Height);
end;


procedure tProgressStatusBar.OnCancelClick(Sender: TObject);
begin
  fShouldCancel := true;
  fsbMain.Panels[0].Text := 'User canceled retrieval';
end;

procedure tProgressStatusBar.OnCancelMouseEnter(Sender: TObject);
begin
  Screen.Cursor := crDefault;
end;

procedure tProgressStatusBar.OnCancelMouseLeave(Sender: TObject);
begin
  Screen.Cursor := fCnRtnCsr;
end;


{$ENDREGION}

{$ENDREGION}


end.
